#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#
from __future__ import annotations

# import json
import pathlib

# import numpy as np
from PySide6.QtCore import (Qt, QSortFilterProxyModel)
from PySide6.QtGui import (QAction, QIcon, )
from PySide6.QtWidgets import (QApplication,  )
from PySide6.QtWidgets import (QDialog, QSplitter, QToolBar,
                               QVBoxLayout, QWidget,)
from quantiphy import Quantity

import revedaEditor.backend.dataDefinitions as ddef
import revedaEditor.backend.libBackEnd as libb
import revedaEditor.backend.libraryMethods as libm
import revedaEditor.backend.libraryModelView as lmview
import revedaEditor.gui.editorViews as edv
import revedaEditor.gui.editorWindow as edw
import revedaEditor.gui.fileDialogues as fd
import revedaEditor.gui.layoutDialogues as ldlg
import revedaEditor.gui.lsw as lsw
from revedaEditor.backend.pdkLoader import importPDKModule
from revedaEditor.scenes.layoutScene import layoutScene

fabproc = importPDKModule('process')
laylyr = importPDKModule('layoutLayers')


class layoutEditor(edw.editorWindow):

    def __init__(self, viewItem: libb.viewItem, libraryDict: dict,
                 libraryView) -> None:
        super().__init__(viewItem, libraryDict, libraryView)
        self.setWindowTitle(f"Layout Editor - {self.cellName} - {self.viewName}")
        self.setWindowIcon(QIcon(":/icons/edLayer-shape.png"))
        self.layoutViews = ["layout", "pcell"]
        self.dbu = fabproc.dbu
        self.majorGrid = fabproc.majorGrid
        self.snapGrid = fabproc.snapGrid
        self.snapTuple = (self.snapGrid, self.snapGrid)
        self.layoutChooser = None
        self.gdsExportDirObj = (self.appMainW.runPath / 'gdsExport' /
                                self.libName / self.cellName)
        self.gdsExportDirObj.mkdir(parents=True, exist_ok=True)
        self._layoutContextMenu()
        # drc error polygons
        self._previousPolygons = []

    def init_UI(self):

        self.resize(1600, 800)
        self._createActions()
        self._createMenuBar()
        self._createToolBars()
        self._addActions()
        self._createTriggers()
        self._createShortcuts()
        # create container to position all widgets
        self.centralW = layoutContainer(self)
        self.setCentralWidget(self.centralW)

    def _createMenuBar(self):
        super()._createMenuBar()
        self.propertyMenu = self.menuEdit.addMenu("Properties")

    def _createActions(self):
        super()._createActions()
        self.renumberInstanceAction = QAction("Renumber Instances", self)
        self.renumberInstanceAction.setToolTip("Renumber Instances")
        cutIcon = QIcon(":/icons/cutter.png")
        self.cutAction = QAction(cutIcon, "Cut Item", self)
        self.cutAction.setToolTip("Cut selected objects")
        self.exportGDSAction = QAction("Export GDS", self)
        self.exportGDSAction.setToolTip("Export GDS from Layout")
        self.exportOASAction = QAction("Export OAS", self)
        self.exportGDSAction.setToolTip("Export OAS from Layout")


    def _addActions(self):
        super()._addActions()
        self.selectMenu.addAction(self.selectDeviceAction)
        self.selectMenu.addAction(self.selectWireAction)
        self.selectMenu.addSeparator()
        self.selectMenu.addAction(self.removeSelectFilterAction)

        self.propertyMenu.addAction(self.objPropAction)
        self.menuEdit.addAction(self.stretchAction)
        self.menuEdit.addAction(self.cutAction)
        self.menuCreate.addAction(self.createInstAction)
        self.menuCreate.addAction(self.createRectAction)
        self.menuCreate.addAction(self.createPathAction)
        self.menuCreate.addAction(self.createPinAction)
        self.menuCreate.addAction(self.createLabelAction)
        self.menuCreate.addAction(self.createViaAction)
        self.menuCreate.addAction(self.createPolygonAction)
        self.menuCreate.addSeparator()
        self.menuCreate.addAction(self.rulerAction)
        self.menuCreate.addAction(self.delRulerAction)
        self.menuTools.addAction(self.renumberInstanceAction)
        self.menuTools.addAction(self.exportGDSAction)
        self.menuTools.addAction(self.exportOASAction)

        # hierarchy submenu
        self.hierMenu = self.menuEdit.addMenu("Hierarchy")
        self.hierMenu.addAction(self.goUpAction)
        self.hierMenu.addAction(self.goDownAction)

        if hasattr(self._app, 'pluginsObj') and hasattr(self._app.pluginsObj, 'applyPluginMenus'):
            self._app.pluginsObj.applyPluginMenus(self)
        if hasattr(self._app, 'pdkConfigObj') and hasattr(self._app.pdkConfigObj, 'applyPDKMenus'):
            self._app.pdkConfigObj.applyPDKMenus(self)

    def _layoutContextMenu(self):
        super()._editorContextMenu()
        self.centralW.scene.itemContextMenu.addAction(self.goDownAction)

    def _createToolBars(self):
        super()._createToolBars()
        self.layoutToolbar = QToolBar("Layout Toolbar", self)
        self.addToolBar(self.layoutToolbar)
        self.layoutToolbar.addAction(self.createInstAction)
        self.layoutToolbar.addAction(self.createRectAction)
        self.layoutToolbar.addAction(self.createPathAction)
        self.layoutToolbar.addAction(self.createPinAction)
        self.layoutToolbar.addAction(self.createLabelAction)
        self.layoutToolbar.addAction(self.createViaAction)
        self.layoutToolbar.addAction(self.createPolygonAction)
        self.layoutToolbar.addSeparator()
        self.layoutToolbar.addAction(self.rulerAction)
        self.layoutToolbar.addAction(self.delRulerAction)
        self.layoutToolbar.addSeparator()
        self.layoutToolbar.addAction(self.goDownAction)
        self.layoutToolbar.addSeparator()
        self.layoutToolbar.addAction(self.removeSelectFilterAction)
        self.layoutToolbar.addAction(self.selectWireAction)
        self.layoutToolbar.addAction(self.selectDeviceAction)

    def _createTriggers(self):
        super()._createTriggers()
        self.createInstAction.triggered.connect(self.createInstClick)
        self.createRectAction.triggered.connect(self.createRectClick)
        self.exportGDSAction.triggered.connect(self.exportGDSClick)
        self.exportOASAction.triggered.connect(self.exportOASClick)
        self.cutAction.triggered.connect(self.cutClick)
        self.createPathAction.triggered.connect(self.createPathClick)
        self.createPinAction.triggered.connect(self.createPinClick)
        self.createLabelAction.triggered.connect(self.createLabelClick)
        self.createViaAction.triggered.connect(self.createViaClick)
        self.createPolygonAction.triggered.connect(self.createPolygonClick)
        self.rulerAction.triggered.connect(self.createRulerClick)
        self.delRulerAction.triggered.connect(self.delRulerClick)
        self.deleteAction.triggered.connect(self.deleteClick)
        self.objPropAction.triggered.connect(self.objPropClick)
        self.goDownAction.triggered.connect(self.goDownClick)
        self.renumberInstanceAction.triggered.connect(self.renumberInstanceClick)

    def _createShortcuts(self):
        super()._createShortcuts()
        self.createRectAction.setShortcut(Qt.Key.Key_R)
        self.createPathAction.setShortcut(Qt.Key.Key_W)
        self.createInstAction.setShortcut(Qt.Key.Key_I)
        self.createPinAction.setShortcut(Qt.Key.Key_P)
        self.createLabelAction.setShortcut(Qt.Key.Key_L)
        self.createViaAction.setShortcut(Qt.Key.Key_V)
        self.createPolygonAction.setShortcut(Qt.Key.Key_G)
        self.stretchAction.setShortcut(Qt.Key.Key_S)
        self.rulerAction.setShortcut(Qt.Key.Key_K)
        self.delRulerAction.setShortcut("Shift+K")
        self.cutAction.setShortcut("Shift+C")

    def cutClick(self):
        self.centralW.scene.editModes.setMode("cutShape")

    def createRectClick(self, s):
        self.centralW.scene.editModes.setMode("drawRect")

    def createRulerClick(self, s):
        self.centralW.scene.editModes.setMode("drawRuler")
        self.messageLine.setText("Click on the first point of the ruler.")

    def delRulerClick(self, s):
        self.messageLine.setText("Deleting all rulers")
        self.centralW.scene.deleteAllRulers()
        self.centralW.scene.editModes.setMode("selectItem")

    def createPathClick(self, s):
        def pathLayerChanged(dlg):
            pathTupleName = dlg.pathLayerCB.currentText()
            pathDefTuple = \
                [item for item in fabproc.processPaths if
                 item.name == pathTupleName][
                    0]
            dlg.pathWidth.setText(pathDefTuple.minWidth.__str__())
            dlg.pathWidthValidator.setRange(pathDefTuple.minWidth,
                                            pathDefTuple.maxWidth)
            dlg.startExtendEdit.setText(str(pathDefTuple.minWidth / 2))
            dlg.endExtendEdit.setText(str(pathDefTuple.minWidth / 2))

        dlg = ldlg.createPathDialogue(self)
        # paths are created on path layers
        processPathNames = [f'{pathTuple.name}' for pathTuple in
                            fabproc.processPaths]
        dlg.pathLayerCB.addItems(processPathNames)
        dlg.pathLayerCB.setCurrentIndex(0)
        defaultPathTuple = fabproc.processPaths[0]
        dlg.pathLayerCB.currentIndexChanged.connect(lambda: pathLayerChanged(dlg))
        dlg.pathWidth.setText(fabproc.processPaths[0].minWidth.__str__())
        dlg.pathWidthValidator.setRange(defaultPathTuple.minWidth,
                                        defaultPathTuple.maxWidth)
        dlg.startExtendEdit.setText(str(fabproc.processPaths[0].minWidth / 2))
        dlg.endExtendEdit.setText(str(fabproc.processPaths[0].minWidth / 2))

        if dlg.exec() == QDialog.DialogCode.Accepted:
            self.centralW.scene.editModes.setMode("drawPath")
            if dlg.manhattanButton.isChecked():
                pathMode = 0
            elif dlg.diagonalButton.isChecked():
                pathMode = 1
            elif dlg.anyButton.isChecked():
                pathMode = 2
            elif dlg.horizontalButton.isChecked():
                pathMode = 3
            elif dlg.verticalButton.isChecked():
                pathMode = 4
            else:
                pathMode = 0
            if dlg.pathWidth.text().strip():
                pathWidth = fabproc.dbu * float(dlg.pathWidth.text().strip())
            else:
                pathWidth = fabproc.dbu * 1.0
            pathName = dlg.pathNameEdit.text()
            pathTupleName = dlg.pathLayerCB.currentText()
            pathTuple = \
                [item for item in fabproc.processPaths if
                 item.name == pathTupleName][
                    0]
            pathLayer = pathTuple.layer
            startExtend = float(dlg.startExtendEdit.text().strip()) * fabproc.dbu
            endExtend = float(dlg.endExtendEdit.text().strip()) * fabproc.dbu
            self.centralW.scene.newPathTuple = ddef.layoutPathTuple(pathName,
                                                                    pathLayer,
                                                                    pathMode,
                                                                    pathWidth,
                                                                    startExtend,
                                                                    endExtend)

    def createPinClick(self):
        dlg = ldlg.createLayoutPinDialog(self)
        pinLayersNames = [f"{item.name} [{item.purpose}]" for item in
                          laylyr.pdkPinLayers]
        textLayersNames = [f"{item.name} [{item.purpose}]" for item in
                           laylyr.pdkTextLayers]
        dlg.pinLayerCB.addItems(pinLayersNames)
        dlg.labelLayerCB.addItems(textLayersNames)

        if self.centralW.scene.newPinTuple is not None:
            dlg.pinLayerCB.setCurrentText(
                f"{self.centralW.scene.newPinTuple.pinLayer.name} "
                f"[{self.centralW.scene.newPinTuple.pinLayer.purpose}]")
        if self.centralW.scene.newLabelTuple is not None:
            dlg.labelLayerCB.setCurrentText(
                f"{self.centralW.scene.newLabelTuple.labelLayer.name} ["
                f"{self.centralW.scene.newLabelTuple.labelLayer.purpose}]")
            dlg.familyCB.setCurrentText(
                self.centralW.scene.newLabelTuple.fontFamily)
            dlg.fontStyleCB.setCurrentText(
                self.centralW.scene.newLabelTuple.fontStyle)
            dlg.labelHeightCB.setCurrentText(
                str(self.centralW.scene.newLabelTuple.fontHeight))
            dlg.labelAlignCB.setCurrentText(
                self.centralW.scene.newLabelTuple.labelAlign)
            dlg.labelOrientCB.setCurrentText(
                self.centralW.scene.newLabelTuple.labelOrient)
        if dlg.exec() == QDialog.DialogCode.Accepted:
            self.centralW.scene.editModes.setMode("drawPin")
            pinName = dlg.pinName.text()
            pinDir = dlg.pinDir.currentText()
            pinType = dlg.pinType.currentText()
            pinLayerName = dlg.pinLayerCB.currentText().split()[0]
            pinLayer = \
                [item for item in laylyr.pdkPinLayers if
                 item.name == pinLayerName][0]
            labelLayerName = dlg.labelLayerCB.currentText().split()[0]
            labelLayer = [item for item in laylyr.pdkTextLayers if
                          item.name == labelLayerName][0]
            fontFamily = dlg.familyCB.currentText()
            fontStyle = dlg.fontStyleCB.currentText()
            labelHeight = float(dlg.labelHeightCB.currentText())
            labelAlign = dlg.labelAlignCB.currentText()
            labelOrient = dlg.labelOrientCB.currentText()
            self.centralW.scene.newPinTuple = ddef.layoutPinTuple(pinName, pinDir,
                                                                  pinType,
                                                                  pinLayer)
            self.centralW.scene.newLabelTuple = ddef.layoutLabelTuple(pinName,
                                                                      fontFamily,
                                                                      fontStyle,
                                                                      labelHeight,
                                                                      labelAlign,
                                                                      labelOrient,
                                                                      labelLayer, )

    def createLabelClick(self):
        dlg = ldlg.createLayoutLabelDialog(self)
        textLayersNames = [f"{item.name} [{item.purpose}]" for item in
                           laylyr.pdkTextLayers]
        dlg.labelLayerCB.addItems(textLayersNames)
        if dlg.exec() == QDialog.DialogCode.Accepted:
            self.centralW.scene.editModes.setMode("addLabel")
            labelName = dlg.labelName.text()
            labelLayerName = dlg.labelLayerCB.currentText().split()[0]
            labelLayer = [item for item in laylyr.pdkTextLayers if
                          item.name == labelLayerName][0]
            fontFamily = dlg.familyCB.currentText()
            fontStyle = dlg.fontStyleCB.currentText()
            fontHeight = dlg.labelHeightCB.currentText()
            labelAlign = dlg.labelAlignCB.currentText()
            labelOrient = dlg.labelOrientCB.currentText()
            self.centralW.scene.newLabelTuple = ddef.layoutLabelTuple(labelName,
                                                                      fontFamily,
                                                                      fontStyle,
                                                                      fontHeight,
                                                                      labelAlign,
                                                                      labelOrient,
                                                                      labelLayer, )

    def createViaClick(self):
        dlg = ldlg.createLayoutViaDialog(self)
        viaLayerNames = [item.name for item in fabproc.processVias]
        dlg.singleViaNamesCB.addItems(viaLayerNames)
        dlg.arrayViaNamesCB.addItems(viaLayerNames)
        dlg.singleViaWidthEdit.setText(str(fabproc.processVias[0].minWidth))
        dlg.singleViaHeightEdit.setText(str(fabproc.processVias[0].minHeight))
        dlg.arrayViaWidthEdit.setText(str(fabproc.processVias[0].minWidth))
        dlg.arrayViaHeightEdit.setText(str(fabproc.processVias[0].minHeight))
        dlg.arrayXspacingEdit.setText(str(fabproc.processVias[0].minSpacing))
        dlg.arrayYspacingEdit.setText(str(fabproc.processVias[0].minSpacing))
        if dlg.exec() == QDialog.DialogCode.Accepted:
            self.centralW.scene.editModes.setMode("addVia")
            self.centralW.scene.addVia = True
            if dlg.singleViaRB.isChecked():
                selViaDefTuple = fabproc.processVias[
                    fabproc.processViaNames.index(
                        dlg.singleViaNamesCB.currentText())]

                singleViaTuple = ddef.singleViaTuple(selViaDefTuple,
                                                     fabproc.dbu * float(
                                                         dlg.singleViaWidthEdit.text().strip()),
                                                     fabproc.dbu * float(
                                                         dlg.singleViaHeightEdit.text().strip()), )
                self.centralW.scene.arrayViaTuple = ddef.arrayViaTuple(
                    singleViaTuple, fabproc.dbu * selViaDefTuple.minSpacing,
                                    fabproc.dbu * selViaDefTuple.minSpacing, 1,
                    1, )
            else:
                selViaDefTuple = \
                    [viaDefTuple for viaDefTuple in fabproc.processVias if
                     viaDefTuple.name == dlg.arrayViaNamesCB.currentText()][0]

                singleViaTuple = ddef.singleViaTuple(selViaDefTuple,
                                                     fabproc.dbu * float(
                                                         dlg.arrayViaWidthEdit.text().strip()),
                                                     fabproc.dbu * float(
                                                         dlg.arrayViaHeightEdit.text().strip()), )
                self.centralW.scene.arrayViaTuple = ddef.arrayViaTuple(
                    singleViaTuple,
                    fabproc.dbu * float(dlg.arrayXspacingEdit.text().strip()),
                    fabproc.dbu * float(dlg.arrayYspacingEdit.text().strip()),
                    int(float(dlg.arrayXNumEdit.text().strip())),
                    int(float(dlg.arrayYNumEdit.text().strip())), )
        else:
            self.centralW.scene.editModes.setMode("selectItem")

    def createPolygonClick(self):
        self.centralW.scene.editModes.setMode("drawPolygon")

    def objPropClick(self, s):
        self.centralW.scene.viewObjProperties()

    def goDownClick(self):
        self.centralW.scene.goDownHier()

    def checkSaveCell(self):
        # need to add checks
        self.centralW.scene.saveLayoutCell(self.file)

    def saveCell(self):
        self.centralW.scene.saveLayoutCell(self.file)

    def loadLayout(self):
        self.logger.info(f'Loading layout from {self.cellName} - {self.viewName}')

        QApplication.setOverrideCursor(Qt.WaitCursor)
        QApplication.processEvents()
        try:
            self.centralW.scene.loadDesign(self.file)
            viewNameTuple = ddef.viewNameTuple(self.libItem.libraryName, self.cellItem.cellName,
                                               self.viewName)
            self.appMainW.openViews[viewNameTuple] = self
        except Exception as e:
            self.logger.error(f"Error during loading layout for {self.cellName}: {e}")
        finally:
            QApplication.restoreOverrideCursor()

    def createInstClick(self, s):
        # create a designLibrariesView
        libraryModel = lmview.layoutViewsModel(self.libraryDict, self.layoutViews)
        if self.layoutChooser is None:
            self.layoutChooser = fd.selectCellViewDialog(self, libraryModel)
            self.layoutChooser.show()
        else:
            self.layoutChooser.raise_()
        if self.layoutChooser.exec() == QDialog.DialogCode.Accepted:
            self.centralW.scene.editModes.setMode("addInstance")
            libItem = libm.getLibItem(libraryModel,
                                      self.layoutChooser.libNamesCB.currentText())
            cellItem = libm.getCellItem(libItem,
                                        self.layoutChooser.cellCB.currentText())
            viewItem = libm.getViewItem(cellItem,
                                        self.layoutChooser.viewCB.currentText())
            # libm.findViewItem(libraryModel, self.layoutChooser.libNamesCB.currentText())
            self.centralW.scene.layoutInstanceTuple = ddef.viewItemTuple(libItem,
                                                                         cellItem,
                                                                         viewItem)

    def _exportLayoutCell(self, export_format="GDS"):
        """Export layout cell in the specified format (GDS or OAS)."""
        # Use appropriate dialogue - both gdsExportDialogue and oasExportDialogue 
        # now inherit from the same base class
        if export_format.upper() == "OAS":
            dlg = fd.oasExportDialogue(self)
        else:
            dlg = fd.gdsExportDialogue(self)
            
        dlg.setWindowTitle(f"Export {export_format.upper()} for {self.cellName}-{self.viewName}")
        dlg.unitEdit.setText(fabproc.gdsUnit.render())
        dlg.precisionEdit.setText(fabproc.gdsPrecision.render())
        dlg.exportPathEdit.setText(str(self.gdsExportDirObj))

        if dlg.exec() == QDialog.DialogCode.Accepted:
            self.gdsExportDir = pathlib.Path(dlg.exportPathEdit.text().strip())
            gdsUnit = Quantity(dlg.unitEdit.text().strip()).real
            gdsPrecision = Quantity(
                dlg.precisionEdit.text().strip()).real
            
            # Call the appropriate export method
            if export_format.upper() == "OAS":
                self.centralW.scene.exportCellOAS(self.gdsExportDir, gdsUnit,
                                                  gdsPrecision, fabproc.dbu)
            else:
                self.centralW.scene.exportCellGDS(self.gdsExportDir, gdsUnit,
                                                  gdsPrecision, fabproc.dbu)

    def exportGDSClick(self):
        """Export layout cell as GDS format."""
        self._exportLayoutCell("GDS")

    def exportOASClick(self):
        """Export layout cell as OAS format."""
        self._exportLayoutCell("OAS")

    def handlePolygonSelection(self, polygons):
        # Remove previous polygons
        for polygon in self._previousPolygons:
            self.centralW.scene.removeItem(polygon)

        # Add new polygons
        for polygon in polygons:
            self.centralW.scene.addItem(polygon)

        # Remember current polygons (store reference)
        self._previousPolygons = polygons

    def renumberInstanceClick(self, s):
        self.centralW.scene.renumberInstances()

    def dispConfigEdit(self):
        import revedaEditor.gui.propertyDialogues as pdlg
        dcd = pdlg.layoutDisplayConfigDialog(self)
        dcd.dbuEntry.setText(str(self.dbu))
        dcd.majorGridEntry.setText(str(self.majorGrid))
        dcd.snapGridEdit.setText(str(self.snapGrid))
        if dcd.exec() == QDialog.DialogCode.Accepted:
            self.configureGridSettings(
                (int(dcd.majorGridEntry.text()), int(dcd.snapGridEdit.text())))
            if dcd.dotType.isChecked():
                self.centralW.view.gridbackg = True
                self.centralW.view.linebackg = False
            elif dcd.lineType.isChecked():
                self.centralW.view.gridbackg = False
                self.centralW.view.linebackg = True
            else:
                self.centralW.view.gridbackg = False
                self.centralW.view.linebackg = False

    def closeEvent(self, event):
        try:
            self.centralW.scene.saveLayoutCell(self.file)
            cellViewNameTuple = ddef.viewNameTuple(self.libName, self.cellName,
                                                   self.viewName)
            self.appMainW.openViews.pop(cellViewNameTuple, None)
        except Exception as e:
            self.appMainW.logger.error(f"Error in closing layout editor window"
                                       f":{self.cellName}-{self.viewName}:{e}")
        finally:
            event.accept()
            super().closeEvent(event)

    # def _createSignalConnections(self):
    #     super()._createSignalConnections()  


class layoutContainer(edw.editorContainer):
    def __init__(self, parent: layoutEditor):
        super().__init__(parent=parent)
        self.editorWindow = parent
        self.scene = layoutScene(self)
        self.view = edv.layoutView(self.scene, self)
        self.lswModel = lsw.layerDataModel(laylyr.pdkAllLayers)
        layerViewTable = lsw.layerViewTable(self, self.lswModel)
        self.lswWidget = lswWindow(layerViewTable)
        self.lswWidget.setMinimumWidth(300)
        self.lswWidget.setMaximumWidth(360)
        self.lswWidget.lswTable.dataSelected.connect(self.selectLayer)
        self.lswWidget.lswTable.layerSelectable.connect(
            self.layerSelectableChange)
        self.lswWidget.lswTable.layerVisible.connect(self.layerVisibleChange)
        self.init_UI()

    def init_UI(self):
        # there could be other widgets in the grid layout, such as edLayer
        # viewer/editor.
        vLayout = QVBoxLayout(self)
        vLayout.setContentsMargins(0, 0, 0, 0)
        vLayout.setSpacing(0)
        self.setContentsMargins(0, 0, 0, 0)
        splitter = QSplitter()
        splitter.setContentsMargins(0, 0, 0, 0)
        splitter.setOrientation(Qt.Orientation.Horizontal)
        splitter.insertWidget(0, self.lswWidget)
        splitter.insertWidget(1, self.view)
        # ratio of first column to second column is 5
        splitter.setStretchFactor(0, 1)
        splitter.setStretchFactor(1, 5)
        vLayout.addWidget(splitter)
        self.setLayout(vLayout)

    def findSelectedLayer(self, layerName: str, layerPurpose: str) -> ddef.layLayer:
        for layer in laylyr.pdkAllLayers:
            if layer.name == layerName and layer.purpose == layerPurpose:
                return layer
        return laylyr.pdkAllLayers[0]

    def selectLayer(self, layerName: str, layerPurpose: str):
        self.scene.selectEdLayer = self.findSelectedLayer(layerName, layerPurpose)

    def layerSelectableChange(self, layerName: str, layerPurpose: str,
                              layerSelectable: bool):
        selectedLayer = self.findSelectedLayer(layerName, layerPurpose)
        selectedLayer.selectable = layerSelectable

        for item in self.scene.items():
            if hasattr(item,
                       "layer") and item.layer == selectedLayer and item.parentItem() is None:
                item.setEnabled(layerSelectable)

    def layerVisibleChange(self, layerName: str, layerPurpose: str,
                           layerVisible: bool):
        selectedLayer = self.findSelectedLayer(layerName, layerPurpose)
        selectedLayer.visible = layerVisible

        for item in self.scene.items():
            if hasattr(item, "layer") and item.layer == selectedLayer:
                item.setVisible(layerVisible)


class LayerFilterProxyModel(QSortFilterProxyModel):
    """Proxy model that matches the search string against layer name OR purpose."""
    def filterAcceptsRow(self, sourceRow, sourceParent):
        pattern = self.filterRegularExpression()
        if pattern.pattern() == "":
            return True
        model = self.sourceModel()
        name_item = model.item(sourceRow, lsw.layerViewTable.columnName)
        purp_item = model.item(sourceRow, lsw.layerViewTable.columnPurpose)
        name = name_item.text() if name_item else ""
        purpose = purp_item.text() if purp_item else ""
        return pattern.match(name).hasMatch() or pattern.match(purpose).hasMatch()


class lswWindow(QWidget):
    def __init__(self, lswTable: lsw.layerViewTable):
        super().__init__()
        self.lswTable = lswTable

        # ensure widget has an opaque background so the underlying scene
        # doesn't show through in the top-left corner
        from PySide6.QtGui import QPalette
        from PySide6.QtWidgets import QLineEdit
        self.setAutoFillBackground(True)
        pal = self.palette()
        pal.setColor(QPalette.Window, pal.color(QPalette.Window))
        self.setPalette(pal)

        layout = QVBoxLayout()
        toolBar = QToolBar()
        avIcon = QIcon(":/icons/eye.png")
        nvIcon = QIcon(":/icons/eye-close.png")
        avAction = QAction(avIcon, "Set Layers Visible", self)
        avAction.setToolTip("Set Layers visible")
        avAction.triggered.connect(self._allLayersVisible)
        nvAction = QAction(nvIcon, "Set Layers Invisible", self)
        nvAction.setToolTip("Set Layers invisible")
        nvAction.triggered.connect(self._noLayersVisible)
        asIcon = QIcon(":/icons/pencil.png")
        nsIcon = QIcon(":/icons/pencil-prohibition.png")
        nsAction = QAction(nsIcon, "Set Layers Selectable", self)
        nsAction.setToolTip("Set layers selectable")
        nsAction.triggered.connect(self._noLayersSelectable)
        asAction = QAction(asIcon, "Set Layers Unselectable", self)
        asAction.setToolTip("Set layers unselectable")
        asAction.triggered.connect(self._allLayersSelectable)

        toolBar.addAction(avAction)
        toolBar.addAction(nvAction)
        toolBar.addAction(asAction)
        toolBar.addAction(nsAction)
        layout.addWidget(toolBar)

        # Search field
        from PySide6.QtWidgets import QLabel, QHBoxLayout
        searchLayout = QHBoxLayout()
        searchLabel = QLabel("Search:")
        self._searchEdit = QLineEdit()
        self._searchEdit.setPlaceholderText("Filter layers by name...")
        self._searchEdit.setToolTip("Type to filter layers by name")
        self._searchEdit.setClearButtonEnabled(True)
        self._searchEdit.setMinimumHeight(24)
        self._searchEdit.setStyleSheet(
            "QLineEdit { color: palette(text); background: palette(base); }"
        )
        searchLayout.addWidget(searchLabel)
        searchLayout.addWidget(self._searchEdit)
        layout.addLayout(searchLayout)

        # Proxy model: filters layer name (col 1) OR purpose (col 2)
        self._proxyModel = LayerFilterProxyModel(self)
        self._proxyModel.setSourceModel(self.lswTable.model())
        self._proxyModel.setFilterCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)
        self.lswTable.setModel(self._proxyModel)
        self._searchEdit.textChanged.connect(self._proxyModel.setFilterFixedString)

        layout.addWidget(self.lswTable)
        self.setLayout(layout)

    def _updateFilteredLayers(self, visible=None, selectable=None):
        """Apply visibility/selectability only to rows currently shown by the filter."""
        sourceModel = self._proxyModel.sourceModel()
        if visible is not None:
            state = Qt.CheckState.Checked if visible else Qt.CheckState.Unchecked
            col = lsw.layerViewTable.columnVisible
        else:
            state = Qt.CheckState.Checked if selectable else Qt.CheckState.Unchecked
            col = lsw.layerViewTable.columnSelectable
        for proxyRow in range(self._proxyModel.rowCount()):
            sourceIndex = self._proxyModel.mapToSource(self._proxyModel.index(proxyRow, col))
            item = sourceModel.item(sourceIndex.row(), col)
            if item is not None:
                item.setCheckState(state)

    def _allLayersVisible(self):
        self._updateFilteredLayers(visible=True)

    def _noLayersVisible(self):
        self._updateFilteredLayers(visible=False)

    def _allLayersSelectable(self):
        self._updateFilteredLayers(selectable=True)

    def _noLayersSelectable(self):
        self._updateFilteredLayers(selectable=False)



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#

import json

from PySide6.QtCore import (Qt, )
from PySide6.QtGui import (QAction, QIcon, QStandardItem, QStandardItemModel, )
from PySide6.QtWidgets import (QApplication, QComboBox, QFormLayout, QGroupBox, QMainWindow,
                               QTableView, QVBoxLayout, QWidget, )

import revedaEditor.backend.dataDefinitions as ddef
import revedaEditor.backend.libBackEnd as libb
import revedaEditor.backend.libraryMethods as libm
import revedaEditor.gui.editFunctions as edf
import revedaEditor.gui.schematicEditor as sced
import revedaEditor.resources.resources  # noqa: F401


class configEditor(QMainWindow):
    def __init__(self, viewItem: libb.viewItem, libraryDict: dict, libraryView):
        super().__init__()
        self.viewItem = viewItem
        self.libraryDict = libraryDict
        self.libraryView = libraryView
        self.configFilePathObj = viewItem.data(Qt.ItemDataRole.UserRole + 2)
        self.cellItem: libb.cellItem = self.viewItem.parent()
        self.libItem: libb.libraryItem = self.cellItem.parent()
        self.libraryName = self.libItem.libraryName
        self.cellName = self.cellItem.cellName
        self.viewName = self.viewItem.viewName

        self._schViewItem = None
        self._schematicEditor = None
        self._configDict = {}
        app = QApplication.instance()
        if app is not None:
            self.appMainW = app.appMainW
            print(self.appMainW.openViews)
        else:
            raise RuntimeError("No QApplication instance found")

        self.setWindowTitle("Edit Config View")
        self.setMinimumSize(600, 700)
        self._createMenuBar()
        self._createActions()
        self._addActions()
        self._createTriggers()

        # Create central widget
        self.centralW = configEditorContainer(self)
        self.setCentralWidget(self.centralW)

    @property
    def configDict(self) -> dict:
        return self._configDict

    @configDict.setter
    def configDict(self, value):
        self._configDict = value
        self._refreshConfigTable()

    @property
    def schViewItem(self) -> libb.viewItem | None:
        return self._schViewItem

    @schViewItem.setter
    def schViewItem(self, value: libb.viewItem):
        self._schViewItem = value

    def loadConfig(self):
        """Load config data from file."""
        try:
            with open(self.viewItem.viewPath) as configFile:
                items = json.load(configFile)

            schematicName = items[1]["reference"]
            self._configDict = items[2]

            self.centralW.libraryNameEdit.setText(self.libraryName)
            self.centralW.cellNameEdit.setText(self.cellName)

            schViewsList = [self.cellItem.child(row).viewName for row in
                            range(self.cellItem.rowCount()) if
                            self.cellItem.child(row).viewType == "schematic"]
            self.centralW.viewNameCB.addItems(schViewsList)
            self.centralW.viewNameCB.setCurrentText(schematicName)

            self.centralW.switchViewsEdit.setText(", ".join(self.appMainW.switchViewList))
            self.centralW.stopViewsEdit.setText(", ".join(self.appMainW.stopViewList))
            configViewNameTuple = ddef.viewNameTuple(self.libraryName, self.cellName,
                                                     self.viewName)
            if self.appMainW.openViews.get(configViewNameTuple, None):
                self.appMainW.openViews[configViewNameTuple].raise_()
            else:
                self.show()
            self._schViewItem = libm.getViewItem(self.cellItem, schematicName)
            schematicViewNameTuple = ddef.viewNameTuple(self.libraryName, self.cellName,
                                                        self._schViewItem.viewName)
            if self.appMainW.openViews.get(schematicViewNameTuple):
                self.editorWindow = self.appMainW.openViews[schematicViewNameTuple]
                self.appMainW.openViews[schematicViewNameTuple].raise_()
            else:
                self.editorWindow = sced.schematicEditor(self._schViewItem,
                                                         self.libraryDict, self.libraryView)
                self.editorWindow.loadSchematic()
                self.editorWindow.show()
            self._refreshConfigTable()

        except Exception as e:
            self.appMainW.logger.error(f'Error loading config: {e}')
            self._configDict = {}

    def _createMenuBar(self):
        self.mainMenu = self.menuBar()
        self.mainMenu.setNativeMenuBar(False)  # for mac
        self.fileMenu = self.mainMenu.addMenu("&File")
        self.editMenu = self.mainMenu.addMenu("&Edit")
        self.status_bar = self.statusBar()
        self.status_bar.showMessage("Ready")

    def _createActions(self):
        updateIcon = QIcon(":/icons/arrow-circle.png")
        self.updateAction = QAction(updateIcon, "Update", self)
        saveIcon = QIcon(":/icons/database--plus.png")
        self.saveAction = QAction(saveIcon, "Save", self)

    def _addActions(self):
        self.fileMenu.addAction(self.updateAction)
        self.fileMenu.addAction(self.saveAction)

    def _createTriggers(self):
        self.updateAction.triggered.connect(self.updateClick)
        self.saveAction.triggered.connect(self.saveCell)

    def updateClick(self):
        newConfigDict = dict()
        if self._schViewItem is not None:
            topSchematicWindow = sced.schematicEditor(self._schViewItem, self.libraryDict,
                                                      self.libraryView, )
        else:
            self.appMainW.logger.error('No schematic view item available')
            return
        topSchematicWindow.loadSchematic()
        topSchematicWindow.createConfigView(self.viewItem, newConfigDict, set())
        self._configDict = newConfigDict
        self._refreshConfigTable()

    def _refreshConfigTable(self):
        self.centralW.confModel = configModel(self.configDict)
        self.centralW.configDictLayout.removeWidget(self.centralW.configViewTable)
        self.centralW.configViewTable = configTable(self.centralW.confModel)
        self.centralW.configDictLayout.addWidget(self.centralW.configViewTable)

    def updateConfigDict(self):
        self.centralW.configViewTable.updateModel()
        self._configDict = dict()
        model = self.centralW.confModel
        for i in range(model.rowCount()):
            viewList = [item.strip() for item in
                        model.itemFromIndex(model.index(i, 3)).text().split(",")]
            self._configDict[model.item(i, 1).text()] = [model.item(i, 0).text(),
                                                         model.item(i, 2).text(),
                                                         viewList, ]

    def saveCell(self):
        # configFilePathObj = self.viewItem.data(Qt.ItemDataRole.UserRole + 2)
        self.updateConfigDict()
        items = list()
        items.insert(0, {"viewName": "config"})
        if self._schViewItem is not None:
            items.insert(1, {"reference": self._schViewItem.viewName})
        else:
            items.insert(1, {"reference": ""})
        items.insert(2, self._configDict)
        with self.configFilePathObj.open(mode="w+") as configFile:
            json.dump(items, configFile, indent=4)

    def checkSaveCell(self):
        self.saveCell()

    def closeEvent(self, event):
        try:
            cellViewNameTuple = ddef.viewNameTuple(self.libraryName, self.cellName,
                                                   self.viewItem.viewName)
            self.appMainW.openViews.pop(cellViewNameTuple, None)
        except Exception as e:
            self.appMainW.logger.error(f"Unexpected error: {e}")
        finally:
            event.accept()
            super().closeEvent(event)


def createNewConfigView(cellItem: libb.cellItem, viewItem: libb.viewItem, dlg,
                        libraryDict: dict, designView, ):
    """Create a new config view from dialog parameters."""
    selectedSchName = dlg.viewNameCB.currentText()
    selectedSchItem = libm.getViewItem(cellItem, selectedSchName)

    schematicWindow = sced.schematicEditor(selectedSchItem, libraryDict, designView, )
    schematicWindow.loadSchematic()
    switchViewList = [viewName.strip() for viewName in dlg.switchViews.text().split(",")]
    stopViewList = [viewName.strip() for viewName in dlg.stopViews.text().split(",")]
    schematicWindow.switchViewList = switchViewList
    schematicWindow.stopViewList = stopViewList

    # clear netlisted cells list
    newConfigDict = dict()  # create an empty newconfig dict
    schematicWindow.createConfigView(viewItem, newConfigDict, set())
    configFilePathObj = viewItem.data(Qt.ItemDataRole.UserRole + 2)
    items = list()
    items.insert(0, {"cellView": "config"})
    items.insert(1, {"reference": selectedSchName})
    items.insert(2, newConfigDict)
    with configFilePathObj.open(mode="w+") as configFile:
        json.dump(items, configFile, indent=4)
    configWindow = configEditor(viewItem, libraryDict, designView)
    configWindow.loadConfig()

    return configWindow


def openConfigEditWindow(schematicItem: libb.viewItem, configItem: libb.viewItem,
                         libraryDict: dict, designView, appMainW, ) -> configEditor:
    """Open an existing config view for editing."""
    schematicName = schematicItem.viewName
    cellItem = schematicItem.parent()
    libItem = cellItem.parent()

    configWindow = configEditor(configItem, libraryDict, designView)
    configWindow.schViewItem = schematicItem
    configWindow.centralW.libraryNameEdit.setText(libItem.libraryName)
    configWindow.centralW.cellNameEdit.setText(cellItem.cellName)

    schViewsList = [cellItem.child(row).viewName for row in range(cellItem.rowCount()) if
                    cellItem.child(row).viewType == "schematic"]
    configWindow.centralW.viewNameCB.addItems(schViewsList)
    configWindow.centralW.viewNameCB.setCurrentText(schematicName)
    configWindow.centralW.switchViewsEdit.setText(", ".join(appMainW.switchViewList))
    configWindow.centralW.stopViewsEdit.setText(", ".join(appMainW.stopViewList))
    configWindow.loadConfig()

    return configWindow


class configEditorContainer(QWidget):
    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        self.mainLayout = QVBoxLayout()
        topCellGroup = QGroupBox("Top Cell")
        topCellLayout = QFormLayout()
        self.libraryNameEdit = edf.longLineEdit()
        topCellLayout.addRow(edf.boldLabel("Library:"), self.libraryNameEdit)
        self.cellNameEdit = edf.longLineEdit()
        topCellLayout.addRow(edf.boldLabel("Cell:"), self.cellNameEdit)
        self.viewNameCB = QComboBox()
        topCellLayout.addRow(edf.boldLabel("View:"), self.viewNameCB)
        topCellGroup.setLayout(topCellLayout)
        self.mainLayout.addWidget(topCellGroup)
        viewGroup = QGroupBox("Switch/Stop Views")
        viewGroupLayout = QFormLayout()
        viewGroup.setLayout(viewGroupLayout)
        self.switchViewsEdit = edf.longLineEdit()
        viewGroupLayout.addRow(edf.boldLabel("View List:"), self.switchViewsEdit)
        self.stopViewsEdit = edf.longLineEdit()
        viewGroupLayout.addRow(edf.boldLabel("Stop List:"), self.stopViewsEdit)
        self.mainLayout.addWidget(viewGroup)
        self.configDictGroup = QGroupBox("Cell View Configuration")
        self.confModel = configModel(self.parent.configDict or {})
        self.configDictLayout = QVBoxLayout()
        self.configViewTable = configTable(self.confModel)
        self.configDictLayout.addWidget(self.configViewTable)
        self.configDictGroup.setLayout(self.configDictLayout)
        self.mainLayout.addWidget(self.configDictGroup)
        self.setLayout(self.mainLayout)


class configModel(QStandardItemModel):
    def __init__(self, configDict: dict):
        row = len(configDict.keys())
        column = 4
        super().__init__(row, column)
        self.setHorizontalHeaderLabels(
            ["Library", "Cell Name", "View Found", "View To ", "Use"])
        for i, (k, v) in enumerate(configDict.items()):
            item = QStandardItem(v[0])
            self.setItem(i, 0, item)
            item = QStandardItem(k)
            self.setItem(i, 1, item)
            item = QStandardItem(v[1])
            self.setItem(i, 2, item)
            item = QStandardItem(", ".join(v[2]))
            self.setItem(i, 3, item)


class configTable(QTableView):
    def __init__(self, model: configModel):
        super().__init__()
        self.configModel = model
        self.setModel(self.configModel)
        self.combos = []
        self.horizontalHeader().setStretchLastSection(True)
        self.setSelectionMode(QTableView.SelectionMode.SingleSelection)
        self.setEditTriggers(QTableView.EditTrigger.NoEditTriggers)

        for row in range(self.configModel.rowCount()):
            combo = QComboBox()
            items = [item.strip() for item in
                     self.configModel.item(row, 3).text().split(",")]
            combo.addItems(items)
            combo.setCurrentText(self.configModel.item(row, 2).text())
            self.setIndexWidget(self.configModel.index(row, 3), combo)
            self.combos.append(combo)

    def updateModel(self):
        for row, combo in enumerate(self.combos):
            self.configModel.setItem(row, 2, QStandardItem(combo.currentText()))



import importlib.resources
import re
from pathlib import Path

import PySide6.QtWebEngineWidgets  # noqa: F401
import markdown
from PySide6.QtCore import QUrl, Qt
from PySide6.QtGui import (QDesktopServices, QAction, QIcon)
from PySide6.QtPrintSupport import (QPrinter, QPrintDialog)
from PySide6.QtWidgets import (QMainWindow, QVBoxLayout, QWidget,
                               QTextBrowser, QDialog, QPushButton, QLabel,
                               QApplication)


def _resolve_docs_path() -> Path:
    """Return the absolute path to the docs directory.

    Search order:
    1. ``revedaEditor/docs/`` via importlib.resources — the correct location
       for an installed wheel where docs is bundled as package data.
    2. Project-root ``docs/`` relative to the running app's basePath — works
       for editable installs and direct source runs.
    3. CWD ``docs/`` — legacy fallback.
    """
    # 1. Installed package data (revedaEditor/docs/)
    try:
        ref = importlib.resources.files("revedaEditor").joinpath("docs")
        candidate = Path(str(ref))
        if candidate.is_dir():
            return candidate
    except (TypeError, ModuleNotFoundError):
        pass

    # 2. Source-tree / editable install via app basePath
    try:
        app = QApplication.instance()
        if app is not None and hasattr(app, "basePath"):
            candidate = Path(app.basePath) / "docs"
            if candidate.is_dir():
                return candidate
    except Exception:
        pass

    # 3. CWD fallback
    return Path("docs")


# from revedaEditor.gui.startThread import startThread

class MarkdownViewer(QTextBrowser):
    def __init__(self, basePath: Path = None):
        super().__init__()
        self.basePath = basePath if basePath is not None else _resolve_docs_path()
        self.currentFile = None

        # Enable link clicking
        self.setOpenLinks(False)  # We'll handle links ourselves
        self.anchorClicked.connect(self.handleLinkClick)

        # Load initial file
        self.loadMarkdownFile("index.md")

    def loadMarkdownFile(self, filename):
        """Load a markdown file and render it"""
        try:
            filePath = self.basePath / filename
            filePath = filePath.resolve()  # Resolve to absolute path

            if not filePath.exists():
                self.setHtml(f"<h1>Error</h1><p>File not found: {filename}</p>")
                return

            with open(filePath, "r", encoding="utf-8") as f:
                text = f.read()

            # Convert markdown to HTML
            html = markdown.markdown(text, extensions=[
                'markdown.extensions.tables',
                'markdown.extensions.fenced_code',
                'markdown.extensions.codehilite',
                'markdown.extensions.toc'
            ])

            # Process images to use absolute file paths
            html = self.processImages(html, filePath.parent)

            # Add styling and set base URL for relative links
            styledHtml = self.addStyling(html)

            # Set the HTML content
            self.setHtml(styledHtml)

            # Update current file reference
            self.currentFile = filePath

        except Exception as e:
            self.setHtml(f"<h1>Error</h1><p>Could not load file: {str(e)}</p>")

    def processImages(self, html, baseDir):
        """Process image tags to use absolute file paths"""

        def replaceImageSrc(match):
            imgTag = match.group(0)
            srcMatch = re.search(r'src=["\']([^"\']+)["\']', imgTag)

            if srcMatch:
                src = srcMatch.group(1)

                # Skip if already absolute URL or file path
                if src.startswith(('http://', 'https://', 'file://', '/')):
                    return imgTag

                # Convert relative path to absolute file path
                absPath = baseDir / src
                if absPath.exists():
                    fileUrl = QUrl.fromLocalFile(str(absPath)).toString()
                    return imgTag.replace(srcMatch.group(0), f'src="{fileUrl}"')
                else:
                    print(f"Warning: Image not found: {absPath}")

            return imgTag

        # Replace all img tags
        return re.sub(r'<img[^>]*>', replaceImageSrc, html)

    def handleLinkClick(self, url):
        """Handle clicking on links in the markdown"""
        urlString = url.toString()

        # Check if it's a markdown file link
        if urlString.endswith('.md') or urlString.endswith('.markdown'):
            # Handle relative paths
            if not urlString.startswith('http'):
                # Calculate the full path
                if self.currentFile:
                    # Resolve relative to current file's directory
                    linkPath = (self.currentFile.parent / urlString).resolve()

                    # Check if the target file exists
                    if linkPath.exists() and linkPath.is_relative_to(self.basePath):
                        # Load the new markdown file
                        relativePath = linkPath.relative_to(self.basePath)
                        self.loadMarkdownFile(str(relativePath))
                        return

                # If relative path resolution failed, try from base path
                baseLinkPath = self.basePath / urlString
                if baseLinkPath.exists():
                    self.loadMarkdownFile(urlString)
                    return

        # For external links or non-markdown files, open in default browser
        if urlString.startswith('http'):
            QDesktopServices.openUrl(url)
        else:
            print(f"Could not resolve link: {urlString}")

    def addStyling(self, htmlContent):
        """Add CSS styling to the HTML"""
        return f"""
        <html>
        <head>
            <style>
                body {{
                    font-family: -apple-system, BlinkMacSystemFont, 'Segoe UI', Roboto, sans-serif;
                    font-size: 12pt;
                    line-height: 1.6;
                    color: #333;
                    max-width: none;
                    margin: 20px;
                    background-color: white;
                }}
                h1, h2, h3, h4, h5, h6 {{
                    color: #2c3e50;
                    margin-top: 12px;
                    margin-bottom: 8px;
                }}
                h1 {{ 
                    border-bottom: 2px solid #eaecef; 
                    padding-bottom: 8px; 
                    font-size: 18pt;
                }}
                h2 {{ 
                    border-bottom: 1px solid #eaecef; 
                    padding-bottom: 6px; 
                    font-size: 14pt;
                }}
                h3 {{ 
                    font-size: 13pt;
                }}
                code {{
                    background-color: #f6f8fa;
                    border-radius: 3px;
                    padding: 2px 4px;
                    font-family: 'Courier New', Consolas, monospace;
                    font-size: 85%;
                    color: #d73a49;
                }}
                pre {{
                    background-color: #f6f8fa;
                    border-radius: 6px;
                    padding: 16px;
                    overflow: auto;
                    line-height: 1.45;
                    border: 1px solid #e1e4e8;
                }}
                pre code {{
                    background-color: transparent;
                    padding: 0;
                    color: #333;
                }}
                blockquote {{
                    border-left: 4px solid #dfe2e5;
                    padding: 0 16px;
                    color: #6a737d;
                    margin: 8px 0;
                    font-style: italic;
                }}
                table {{
                    border-collapse: collapse;
                    width: 100%;
                    margin: 8px 0;
                }}
                th, td {{
                    border: 1px solid #dfe2e5;
                    padding: 8px 12px;
                    text-align: left;
                }}
                th {{
                    background-color: #f6f8fa;
                    font-weight: bold;
                }}
                a {{
                    color: #0366d6;
                    text-decoration: none;
                }}
                a:hover {{
                    text-decoration: underline;
                    color: #0256cc;
                }}
                ul, ol {{
                    padding-left: 24px;
                    margin: 8px 0;
                }}
                li {{
                    margin: 4px 0;
                }}
                hr {{
                    border: none;
                    border-top: 1px solid #eaecef;
                    margin: 24px 0;
                }}
                p {{
                    margin: 8px 0;
                }}
                img {{
                    max-width: 100%;
                    height: auto;
                    margin: 4px 0;
                }}
                figure {{
                    margin: 4px 0;
                }}
                figcaption {{
                    margin: 2px 0;
                }}
            </style>
        </head>
        <body>
            {htmlContent}
        </body>
        </html>
        """


class helpBrowser(QMainWindow):
    def __init__(self, _parent):
        super().__init__(parent=_parent)
        self.appMainW = _parent
        self.setWindowTitle("Revolution EDA Documentation Viewer")
        self.setGeometry(100, 100, 1000, 700)

        # Navigation history
        self.history = []
        self.historyIndex = -1

        # Create toolbar
        toolbar = self.addToolBar("Navigation")

        # Home button
        homeAction = QAction(QIcon(":/icons/home.png"), "Home", self)
        homeAction.triggered.connect(self.goHome)
        toolbar.addAction(homeAction)

        # Back button
        self.backAction = QAction(QIcon(":/icons/arrow-180.png"), "Back", self)
        self.backAction.triggered.connect(self.goBack)
        self.backAction.setEnabled(False)
        toolbar.addAction(self.backAction)
        self.backAction.setToolTip('Go back')

        # Forward button
        self.forwardAtion = QAction(QIcon(":/icons/arrow.png"), "Forward", self)
        self.forwardAtion.triggered.connect(self.goForward)
        self.forwardAtion.setEnabled(False)
        toolbar.addAction(self.forwardAtion)

        toolbar.addSeparator()

        # Print button
        printAction = QAction(QIcon(":/icons/printer--arrow.png"), "Print", self)
        printAction.triggered.connect(self.printDocument)
        toolbar.addAction(printAction)

        # Create central widget
        centralW = QWidget()
        self.setCentralWidget(centralW)
        layout = QVBoxLayout(centralW)

        # Create markdown viewer
        self.markdownViewer = MarkdownViewer()
        self.markdownViewer.anchorClicked.connect(self.handleNavigation)
        layout.addWidget(self.markdownViewer)

        # Add initial page to history
        self.addToHistory("index.md")

    def addToHistory(self, filename):
        """Add a page to navigation history"""
        # Remove forward history if we're not at the end
        if self.historyIndex < len(self.history) - 1:
            self.history = self.history[:self.historyIndex + 1]

        self.history.append(filename)
        self.historyIndex = len(self.history) - 1
        self.updateNavigationButtons()

    def updateNavigationButtons(self):
        """Update the enabled state of navigation buttons"""
        self.backAction.setEnabled(self.historyIndex > 0)
        self.forwardAtion.setEnabled(self.historyIndex < len(self.history) - 1)

    def goHome(self):
        """Navigate to home page"""
        self.markdownViewer.loadMarkdownFile("index.md")
        self.addToHistory("index.md")

    def goBack(self):
        """Navigate back in history"""
        if self.historyIndex > 0:
            self.historyIndex -= 1
            filename = self.history[self.historyIndex]
            self.markdownViewer.loadMarkdownFile(filename)
            self.updateNavigationButtons()

    def goForward(self):
        """Navigate forward in history"""
        if self.historyIndex < len(self.history) - 1:
            self.historyIndex += 1
            filename = self.history[self.historyIndex]
            self.markdownViewer.loadMarkdownFile(filename)
            self.updateNavigationButtons()

    def handleNavigation(self, url):
        """Handle navigation and update history"""
        urlString = url.toString()
        if urlString.endswith('.md') or urlString.endswith('.markdown'):
            self.addToHistory(urlString)
        self.markdownViewer.handleLinkClick(url)

    def printDocument(self):
        """Print the current document"""
        printer = QPrinter(QPrinter.HighResolution)
        dialog = QPrintDialog(printer, self)

        if dialog.exec() == QPrintDialog.Accepted:
            self.markdownViewer.print_(printer)


class aboutDialog(QDialog):
    def __init__(self, parent):
        super().__init__(parent)

        self.setWindowTitle("About Revolution EDA")
        self.setGeometry(100, 100, 400, 200)
        self.setWindowFlags(Qt.Dialog | Qt.WindowStaysOnTopHint)
        self.setMinimumWidth(300)

        layout = QVBoxLayout()

        # Add information about your application using rich text
        aboutLabel = QLabel(
            "<h2>Revolution EDA</h2>"
            "<p><strong>Version:</strong> 0.8.10</p>"
            "<p><strong>Copyright: Revolution Semiconductor</strong> © 2026</p>"
            "<p><strong>License:</strong> Mozilla Public License 2.0 amended with Commons Clause</p>"
            "<p><strong> Website:</strong> <a href='https://reveda.eu'>Revolution EDA</a></p>"
            "<p><strong> GitHub:</strong> "
            "<a href='https://github.com/eskiyerli/revolution-eda'>Revolution "
            "EDA GitHub Repository</a></p>"
        )
        aboutLabel.setOpenExternalLinks(True)  # Allow clickable links
        layout.addWidget(aboutLabel)
        layout.addSpacing(20)

        # Add a "Close" button
        closeButton = QPushButton("Close")
        closeButton.clicked.connect(self.accept)
        layout.addWidget(closeButton)

        self.setLayout(layout)



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#

# from hashlib import new
import json
# from hashlib import new
import pathlib

import orjson
from copy import deepcopy
from typing import List, Dict

# import numpy as np
from PySide6.QtCore import (
    QLineF,
    QPoint,
    QPointF,
    Qt,
    Signal,
)
from PySide6.QtGui import (
    QColor,
    QGuiApplication,
    QPen,
)
from PySide6.QtWidgets import (
    QDialog,
    QGraphicsLineItem,
    QGraphicsSceneMouseEvent,
)

import revedaEditor.backend.dataDefinitions as ddef
import revedaEditor.backend.undoStack as us
import revedaEditor.common.labels as lbl
import revedaEditor.common.shapes as shp  # import the shapes
import revedaEditor.fileio.loadJSON as lj
import revedaEditor.fileio.symbolEncoder as symenc
import revedaEditor.gui.alignItems as alg
import revedaEditor.gui.propertyDialogues as pdlg
from revedaEditor.backend.pdkLoader import importPDKModule
from revedaEditor.scenes.editorScene import editorScene

symlyr = importPDKModule('symLayers')


# noinspection PyUnresolvedReferences
class symbolScene(editorScene):
    """
    Scene for Symbol editor.
    """
    alignLineFinished = Signal(shp.alignLine)

    def __init__(self, parent):
        super().__init__(parent)
        # drawing modes
        self.editModes = ddef.symbolModes(
            selectItem=True,
            deleteItem=False,
            moveItem=False,
            copyItem=False,
            rotateItem=False,
            changeOrigin=False,
            panView=False,
            zoomView=False,
            drawPin=False,
            drawArc=False,
            drawRect=False,
            drawLine=False,
            addLabel=False,
            drawCircle=False,
            drawPolygon=False,
            stretchItem=False,
            alignItems=False,
        )

        self.messages = {
            "selectItem": "Select Item.",
            "deleteItem": "Click on item to delete.",
            "moveItem": "Click and drag to move item.",
            "copyItem": "Click on item to copy.",
            "rotateItem": "Click on item to rotate.",
            "changeOrigin": "Click to set new origin.",
            "panView": "Click and drag to pan view.",
            'zoomView': 'Draw rectangle to zoom in',
            "drawPin": "Add a pin.",
            "drawArc": "Click for the first point of Arc.",
            "drawRect": "Click for the first point of Rectangle.",
            "drawLine": "Click for the first point of Line.",
            "addLabel": "Adding a label.",
            "drawCircle": "Click for the centre of Circle.",
            "drawPolygon": "Click for the first point of Polygon.",
            "stretchItem": "Click and drag to stretch item.",
            "alignItems": "Select Align Option."
        }

        self.symbolShapes = ["line", "arc", "rect", "circle", "pin", "label", "polygon"]
        self.attributeList = []  # list of symbol attributes
        self.origin = QPoint(0, 0)
        # some default attributes

        self.pinName = ""
        self.pinType = shp.symbolPin.pinTypes[0]
        self.pinDir = shp.symbolPin.pinDirs[0]
        self.labelDefinition = ""
        self.labelType = lbl.symbolLabel.labelTypes[0]
        self.labelOrient = lbl.symbolLabel.labelOrients[0]
        self.labelAlignment = lbl.symbolLabel.labelAlignments[0]
        self.labelUse = lbl.symbolLabel.labelUses[0]
        self.labelVisible = False
        self.labelHeight = "12"
        self.labelOpaque = True
        self.newPin = None
        self.newLine = None
        self.newRect = None
        self.newCircle = None
        self.newArc = None
        self.newLabel = None
        self._newPolygon = None
        self.polygonGuideLine = None

        self.alignLineFinished.connect(alg.alignToLine)

    @property
    def drawMode(self):
        return any(
            (
                self.editModes.drawPin,
                self.editModes.drawArc,
                self.editModes.drawLine,
                self.editModes.drawRect,
                self.editModes.drawCircle,
                self.editModes.drawPolygon,
            )
        )

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mouseMoveEvent(event)
        self.mouseMoveLoc = self.snapToGrid(event.scenePos().toPoint())

        # Update message only when needed
        message = None
        if self.editModes.drawLine and self.newLine:
            message = "Release mouse on the end point"
            self.newLine.end = self.mouseMoveLoc
        elif self.editModes.drawPin and self.newPin:
            message = "Place pin"
            self.newPin.setPos(self.mouseMoveLoc - self.mouseReleaseLoc)
        elif self.editModes.drawCircle and self.newCircle:
            message = "Extend Circle"
            # Optimize circle radius calculation
            dx = self.mouseMoveLoc.x() - self.newCircle.centre.x()
            dy = self.mouseMoveLoc.y() - self.newCircle.centre.y()
            self.newCircle.radius = (dx * dx + dy * dy) ** 0.5
        elif self.editModes.drawRect and self.newRect:
            message = "Click to finish the rectangle"
            self.newRect.end = self.mouseMoveLoc
        elif self.editModes.drawArc and self.newArc:
            message = "Extend Arc"
            self.newArc.end = self.mouseMoveLoc
        elif self.editModes.addLabel and self.newLabel:
            message = "Place Label"
            self.newLabel.setPos(self.mouseMoveLoc)
        elif (self.editModes.drawPolygon and
              self._newPolygon and
              self.polygonGuideLine):
            message = "Add another point to Polygon"
            self.polygonGuideLine.setLine(
                QLineF(self._newPolygon.points[-1], self.mouseMoveLoc)
            )
        elif self.newAlignLine and self.editModes.alignItems:
            self.newAlignLine.draftLine = QLineF(
                self.newAlignLine.draftLine.p1(), self.mouseMoveLoc)
        if message:
            self.editorWindow.messageLine.setText(message)

        self.statusLine.showMessage(
            f"Cursor Position: {(self.mouseMoveLoc - self.origin).toTuple()}")

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mouseReleaseEvent(event)

    def _handleMouseRelease(self, mousePos: QPoint, button: Qt.MouseButton) -> None:
        modifiers = QGuiApplication.keyboardModifiers()
        try:
        #    if self.editModes.changeOrigin:
        #         self.origin = mousePos
            if self.editModes.drawLine:
                if self.newLine:
                    if self.newLine.length <= 1:
                        self.undoStack.removeLastCommand()
                    self.newLine = None
                self.newLine = self.lineDraw(mousePos, mousePos)
                self.newLine.setSelected(True)
            elif self.editModes.drawCircle:
                if self.newCircle:
                    if self.newCircle.radius <= 1:
                        self.undoStack.removeLastCommand()
                    self.newCircle = None

                self.newCircle = self.circleDraw(mousePos, mousePos)
                self.newCircle.setSelected(True)
            elif self.editModes.drawPin:
                if self.newPin:
                    self.newPin = None
                self.newPin = self.pinDraw(mousePos)
                self.newPin.setSelected(True)
            elif self.editModes.drawRect:
                if self.newRect:
                    if self.newRect.width <= 1 or self.newRect.height <= 1:
                        self.undoStack.removeLastCommand()
                    self.newRect = None
                self.newRect = self.rectDraw(mousePos, mousePos)
                self.newRect.setSelected(True)
            elif self.editModes.drawArc:
                if self.newArc:
                    if self.newArc.width <= 1 or self.newArc.height <= 1:
                        self.undoStack.removeLastCommand()
                    self.newArc = None
                self.newArc = self.arcDraw(mousePos, mousePos)
                self.newArc.setSelected(True)
            elif self.editModes.addLabel:
                if self.newLabel:
                    self.newLabel = None
                self.newLabel = self.labelDraw(
                    mousePos,
                    self.labelDefinition,
                    self.labelType,
                    self.labelHeight,
                    self.labelAlignment,
                    self.labelOrient,
                    self.labelUse,
                )
                self.newLabel.setSelected(True)
            elif self.editModes.drawPolygon:
                if self._newPolygon:
                    self._newPolygon.addPoint(mousePos)
                else:
                    self._newPolygon, self.polygonGuideLine = self.startPolygon(mousePos)
            elif self.editModes.rotateItem:
                self.rotateSelectedItems(mousePos)
            elif self.editModes.alignItems:
                self._handleAlignItemLine(mousePos)
            self.messageLine.setText(self.messages.get(self.editModes.mode(), ""))
        except Exception as e:
            self.logger.error(f"Error in Mouse Release Event: {e} ")

    def mouseDoubleClickEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mouseDoubleClickEvent(event)
        try:
            self.finishPolygon(event)
        except Exception as e:
            self.logger.error(f"Error in mouse Double Click Event: {e}")

    def lineDraw(self, start: QPoint, current: QPoint):
        line = shp.symbolLine(start, current)
        self.addUndoStack(line)
        # self.addItem(line)
        # undoCommand = us.addShapeUndo(self, line)
        # self.undoStack.push(undoCommand)
        return line

    def rectDraw(self, start: QPoint, end: QPoint):
        """
        Draws a rectangle on the scene
        """
        rect = shp.symbolRectangle(start, end)
        self.addUndoStack(rect)
        # undoCommand = us.addShapeUndo(self, rect)
        # self.undoStack.push(undoCommand)
        return rect

    def circleDraw(self, start: QPoint, end: QPoint):
        """
        Draws a circle on the scene
        """
        circle = shp.symbolCircle(start, end)

        self.addUndoStack(circle)
        return circle

    def arcDraw(self, start: QPoint, end: QPoint):
        """
        Draws an arc inside the rectangle defined by start and end points.
        """
        arc = shp.symbolArc(start, end)

        self.addUndoStack(arc)
        return arc

    def pinDraw(self, current):
        pin = shp.symbolPin(current, self.pinName, self.pinDir, self.pinType)
        # self.addItem(pin)
        self.addUndoStack(pin)
        return pin

    def labelDraw(
            self,
            start,
            labelDefinition,
            labelType,
            labelHeight,
            labelAlignment,
            labelOrient,
            labelUse,
    ):
        label = lbl.symbolLabel(
            start,
            labelDefinition,
            labelType,
            labelHeight,
            labelAlignment,
            labelOrient,
            labelUse,
        )
        label.labelVisible = self.labelOpaque
        label.labelDefs()
        label.setOpacity(1)
        self.addUndoStack(label)
        return label

    def addProperty(self, labelDefinition: str):
        # Extract core expression between [@ and ]
        if labelDefinition.startswith("[@") and labelDefinition.endswith("]"):
            core = labelDefinition[2:-1]
            # Split into components
            parts = core.split(":")
            instProp = parts[0]

    def startPolygon(self, startLoc: QPoint):
        newPolygon = shp.symbolPolygon([startLoc])
        self.addUndoStack(newPolygon)
        # Create guide line
        guideLine = QLineF(startLoc, startLoc)
        polygonGuideLine = QGraphicsLineItem(guideLine)
        polygonGuideLine.setPen(QPen(QColor(255, 255, 0), 1, Qt.DashLine))
        self.addUndoStack(polygonGuideLine)
        return newPolygon, polygonGuideLine

    def finishPolygon(self, event):
        if (hasattr(event, 'button') and event.button() == Qt.MouseButton.LeftButton
                and self.editModes.drawPolygon and self._newPolygon):
            self.editModes.setMode("selectItem")
            self._newPolygon = None
            self.removeItem(self.polygonGuideLine)
            self.polygonGuideLine = None
            self.editorWindow.messageLine.setText('Select Item.')

    def _handleAlignItemLine(self, eventLoc: QPoint) -> None:

        if self.newAlignLine:
            if not self.newAlignLine.draftLine.isNull():
                self.alignLineFinished.emit(self.newAlignLine)
        else:
            from PySide6.QtWidgets import (QApplication, )
            alignDialogs = [w for w in QApplication.topLevelWidgets() if isinstance(w,
                                                                                    alg.alignItemsDialogue) and w.isVisible() and w.scene == self]
            if not alignDialogs:
                self.logger.warning("No align items dialogue found")
                return
            alignDlg = alignDialogs[0]
            if alignDlg.horizontalAlignButton.isChecked():
                self.newAlignLine = shp.alignLine(QLineF(eventLoc, eventLoc), 1,
                                                  0)  # horizontal
            else:
                self.newAlignLine = shp.alignLine(QLineF(eventLoc, eventLoc), 1,
                                                  1)
            self.addUndoStack(self.newAlignLine)

    def moveBySelectedItems(self):
        if self.selectedItems():
            dlg = pdlg.moveByDialogue(self.editorWindow)
            dlg.xEdit.setText("0")
            dlg.yEdit.setText("0")
            if dlg.exec() == QDialog.DialogCode.Accepted:
                for item in self.selectedItems():
                    item.moveBy(
                        self.snapToBase(float(dlg.xEdit.text()), self.snapTuple[0]),
                        self.snapToBase(float(dlg.yEdit.text()), self.snapTuple[1]),
                    )
            self.editorWindow.messageLine.setText(
                f"Moved items by {dlg.xEdit.text()} and {dlg.yEdit.text()}"
            )
            self.editModes.setMode("selectItem")

    def itemProperties(self):
        """
        When item properties is queried.
        """
        # Define the mapping of types to update methods
        UPDATE_METHODS = {
            shp.symbolRectangle: self.updateSymbolRectangle,
            shp.symbolCircle: self.updateSymbolCircle,
            shp.symbolArc: self.updateSymbolArc,
            shp.symbolLine: self.updateSymbolLine,
            shp.symbolPin: self.updateSymbolPin,
            lbl.symbolLabel: self.updateSymbolLabel,
            shp.symbolPolygon: self.updateSymbolPolygon
        }

        # Filter selected items and update them
        for item in (item for item in self.selectedItems() if item.parentItem() is None):
            updateMethod = UPDATE_METHODS.get(type(item))
            if updateMethod:
                updateMethod(item)

    def updateSymbolRectangle(self, item):
        queryDlg = pdlg.rectPropertyDialog(self.editorWindow)
        [left, top, width, height] = item.rect.getRect()
        sceneTopLeftPoint = item.mapToScene(QPoint(left, top))
        queryDlg.rectLeftLine.setText(str(sceneTopLeftPoint.x()))
        queryDlg.rectTopLine.setText(str(sceneTopLeftPoint.y()))
        queryDlg.rectWidthLine.setText(str(width))  # str(width))
        queryDlg.rectHeightLine.setText(str(height))  # str(height))
        if queryDlg.exec() == QDialog.DialogCode.Accepted:
            newRectItem = shp.symbolRectangle(QPoint(0, 0), QPoint(0, 0))
            newRectItem.left = self.snapToBase(
                float(queryDlg.rectLeftLine.text()), self.snapTuple[0]
            )
            newRectItem.top = self.snapToBase(
                float(queryDlg.rectTopLine.text()), self.snapTuple[1]
            )
            newRectItem.width = self.snapToBase(
                float(queryDlg.rectWidthLine.text()), self.snapTuple[0]
            )
            newRectItem.height = self.snapToBase(
                float(queryDlg.rectHeightLine.text()), self.snapTuple[1]
            )

            self.undoStack.push(us.addDeleteShapeUndo(self, newRectItem, item))

    def updateSymbolCircle(self, item):
        queryDlg = pdlg.circlePropertyDialog(self.editorWindow)
        centre = item.mapToScene(item.centre).toTuple()
        radius = item.radius
        queryDlg.centerXEdit.setText(str(centre[0]))
        queryDlg.centerYEdit.setText(str(centre[1]))
        queryDlg.radiusEdit.setText(str(radius))
        if queryDlg.exec() == QDialog.DialogCode.Accepted:
            newCircleItem = shp.symbolCircle(QPoint(0, 0), QPoint(0, 0))
            centerX = self.snapToBase(
                float(queryDlg.centerXEdit.text()), self.snapTuple[0]
            )
            centerY = self.snapToBase(
                float(queryDlg.centerYEdit.text()), self.snapTuple[1]
            )
            newCircleItem.centre = QPoint(centerX, centerY)
            radius = self.snapToBase(
                float(queryDlg.radiusEdit.text()), self.snapTuple[0]
            )
            newCircleItem.radius = radius
            self.undoStack.push(us.addDeleteShapeUndo(self, newCircleItem, item))

    def updateSymbolLine(self, item):
        queryDlg = pdlg.linePropertyDialog(self.editorWindow)
        sceneLineStartPoint = item.mapToScene(item.start).toPoint()
        sceneLineEndPoint = item.mapToScene(item.end).toPoint()
        queryDlg.startXLine.setText(str(sceneLineStartPoint.x()))
        queryDlg.startYLine.setText(str(sceneLineStartPoint.y()))
        queryDlg.endXLine.setText(str(sceneLineEndPoint.x()))
        queryDlg.endYLine.setText(str(sceneLineEndPoint.y()))
        if queryDlg.exec() == QDialog.DialogCode.Accepted:
            startX = self.snapToBase(
                float(queryDlg.startXLine.text()), self.snapTuple[0]
            )
            startY = self.snapToBase(
                float(queryDlg.startYLine.text()), self.snapTuple[1]
            )
            endX = self.snapToBase(
                float(queryDlg.endXLine.text()), self.snapTuple[0]
            )
            endY = self.snapToBase(
                float(queryDlg.endYLine.text()), self.snapTuple[1]
            )
            newLine = shp.symbolLine(QPoint(startX, startY), QPoint(endX, endY))
            self.undoStack.push(us.addDeleteShapeUndo(self, newLine, item))

    def updateSymbolArc(self, item):
        queryDlg = pdlg.arcPropertyDialog(self.editorWindow)
        sceneStartPoint = item.mapToScene(item.start).toPoint()
        queryDlg.startXEdit.setText(str(sceneStartPoint.x()))
        queryDlg.startYEdit.setText(str(sceneStartPoint.y()))
        queryDlg.widthEdit.setText(str(item.width))
        queryDlg.heightEdit.setText(str(item.height))
        arcType = item.arcType
        if queryDlg.exec() == QDialog.DialogCode.Accepted:
            startX = int(float(queryDlg.startXEdit.text()))
            startY = int(float(queryDlg.startYEdit.text()))
            start = self.snapToGrid(QPoint(startX, startY))
            width = int(float(queryDlg.widthEdit.text()))
            height = int(float(queryDlg.heightEdit.text()))
            end = start + QPoint(width, height)
            newArc = shp.symbolArc(start, end)
            newArc.arcType = arcType

            self.undoStack.push(us.addDeleteShapeUndo(self, newArc, item))
            newArc.height = height

    def updateSymbolLabel(self, item):
        queryDlg = pdlg.labelPropertyDialog(self.editorWindow)
        queryDlg.labelDefinition.setText(str(item.labelDefinition))
        queryDlg.labelHeightEdit.setText(str(item.labelHeight))
        queryDlg.labelAlignCombo.setCurrentText(item.labelAlign)
        queryDlg.labelOrientCombo.setCurrentText(item.labelOrient)
        queryDlg.labelUseCombo.setCurrentText(item.labelUse)
        if item.labelVisible:
            queryDlg.labelVisiCombo.setCurrentText("Yes")
        else:
            queryDlg.labelVisiCombo.setCurrentText("No")
        if item.labelType == "Normal":
            queryDlg.normalType.setChecked(True)
        elif item.labelType == "NLPLabel":
            queryDlg.NLPType.setChecked(True)
        elif item.labelType == "PyLabel":
            queryDlg.pyLType.setChecked(True)
        sceneStartPoint = item.pos()
        queryDlg.labelXLine.setText(str(sceneStartPoint.x()))
        queryDlg.labelYLine.setText(str(sceneStartPoint.y()))
        if queryDlg.exec() == QDialog.DialogCode.Accepted:
            startX = int(float(queryDlg.labelXLine.text()))
            startY = int(float(queryDlg.labelYLine.text()))
            start = self.snapToGrid(QPoint(startX, startY))
            labelDefinition = queryDlg.labelDefinition.text()
            labelHeight = int(float(queryDlg.labelHeightEdit.text()))
            labelAlign = queryDlg.labelAlignCombo.currentText()
            labelOrient = queryDlg.labelOrientCombo.currentText()
            labelUse = queryDlg.labelUseCombo.currentText()
            labelType = lbl.symbolLabel.labelTypes[0]
            if queryDlg.NLPType.isChecked():
                labelType = lbl.symbolLabel.labelTypes[1]
            elif queryDlg.pyLType.isChecked():
                labelType = lbl.symbolLabel.labelTypes[2]
            newLabel = lbl.symbolLabel(
                start,
                labelDefinition,
                labelType,
                labelHeight,
                labelAlign,
                labelOrient,
                labelUse,
            )
            newLabel.labelVisible = (
                    queryDlg.labelVisiCombo.currentText() == "Yes"
            )
            newLabel.labelDefs()
            newLabel.setOpacity(1)
            self.undoStack.push(us.addDeleteShapeUndo(self, newLabel, item))
        else:
            self.addItem(item)

    def updateSymbolPin(self, item):
        queryDlg = pdlg.pinPropertyDialog(self.editorWindow)
        queryDlg.pinName.setText(str(item.pinName))
        queryDlg.pinType.setCurrentText(item.pinType)
        queryDlg.pinDir.setCurrentText(item.pinDir)
        sceneStartPoint = item.mapToScene(item.start).toPoint()
        queryDlg.pinXLine.setText(str(sceneStartPoint.x()))
        queryDlg.pinYLine.setText(str(sceneStartPoint.y()))
        if queryDlg.exec() == QDialog.DialogCode.Accepted:
            sceneStartX = int(float(queryDlg.pinXLine.text()))
            sceneStartY = int(float(queryDlg.pinYLine.text()))
            start = self.snapToGrid(
                QPoint(sceneStartX, sceneStartY))
            pinName = queryDlg.pinName.text()
            pinType = queryDlg.pinType.currentText()
            pinDir = queryDlg.pinDir.currentText()
            newPin = shp.symbolPin(start, pinName, pinDir, pinType)
            self.undoStack.push(us.addDeleteShapeUndo(self, newPin, item))

    def updateSymbolPolygon(self, item):
        pointsTupleList = [(point.x(), point.y()) for point in item.points]
        queryDlg = pdlg.symbolPolygonProperties(
            self.editorWindow, pointsTupleList
        )
        if queryDlg.exec() == QDialog.DialogCode.Accepted:
            tempPoints = []
            for i in range(queryDlg.tableWidget.rowCount()):
                xcoor = queryDlg.tableWidget.item(i, 1).text()
                ycoor = queryDlg.tableWidget.item(i, 2).text()
                if xcoor != "" and ycoor != "":
                    tempPoints.append(QPointF(float(xcoor), float(ycoor)))
            newPolygon = shp.symbolPolygon(tempPoints)
            self.undoStack.push(us.addDeleteShapeUndo(self, newPolygon, item))

    def loadDesign(self, filePathObj: pathlib.Path) -> None:
        try:
            with filePathObj.open("rb") as file:
                decodedData = orjson.loads(file.read())
            self.blockSignals(True)
            with self.measureDuration():
                viewDict, gridSettings, *itemData = decodedData
                if viewDict.get("viewType") != "symbol":
                    raise Exception("Not a symbol file!")
                if gridSettings and gridSettings.get("snapGrid"):
                    self.editorWindow.configureGridSettings(decodedData[1].get(
                        "snapGrid", (self.majorGrid,
                                     self.snapGrid)))
                self.attributeList = []
                self.createSymbolItems(itemData)
            self.itemsRef = set(self.items())
        except (orjson.JSONDecodeError, FileNotFoundError) as e:
            self.logger.error(f"File error while loading symbol: {e}")
            self.attributeList = []
            self.clear()
            return
        except KeyError as e:
            self.logger.error(f"Invalid symbol format - missing key: {e}")
            self.attributeList = []
            self.clear()
            return
        except Exception as e:
            self.logger.error(f"Unexpected error loading symbol: {e}")
            self.attributeList = []
            self.clear()
            return
        finally:
            self.blockSignals(False)
            self.update()

    def createSymbolItems(self, itemsList: List[Dict]):
        factory = lj.symbolItems(self)
        for itemDict in itemsList:
            if itemDict:
                itemType = itemDict.get("type", None)
                if itemType in self.symbolShapes:
                    itemShape = factory.create(itemDict)
                    if itemType == "label":
                        itemShape.setOpacity(1)
                    self.addItem(itemShape)
                    continue
                if itemType == "attr":
                    self.attributeList.append(factory.createSymbolAttribute(itemDict))

    def saveSymbolCell(self, fileName: pathlib.Path) -> bool:
        """
        Save symbol cell to file.

        Args:
            fileName: Path to save the symbol cell

        Returns:
            bool: True if save successful, False otherwise
        """

        try:
            self.itemsRefSet = set(self.items())
            # Filter items and process labels in one pass

            symbolClasses = (shp.symbolPolygon, shp.symbolRectangle, shp.symbolArc,
                             shp.symbolLine, shp.symbolPin, lbl.symbolLabel,
                             shp.symbolCircle, shp.text)
            sceneItems = [item for item in self.items() if isinstance(item,
                                                                      symbolClasses)]
            # Build save data
            save_data = [
                {"viewType": "symbol"},
                {"snapGrid": (self.majorGrid, self.snapGrid)},
                *sceneItems,
                *getattr(self, "attributeList", [])
            ]

            # Write to file
            with fileName.open("w") as f:
                json.dump(save_data, f, cls=symenc.symbolEncoder, indent=4)

            self.undoStack.clear()
            return True

        except Exception as e:
            self.logger.error(f"Symbol save error: {e}")
            return False

    def viewSymbolProperties(self):
        """
        View symbol properties dialog.
        """
        # copy symbol attribute list to another list by deepcopy to be safe
        attributeListCopy = deepcopy(self.attributeList)
        symbolPropDialogue = pdlg.symbolLabelsDialogue(
            self.editorWindow, self.items(), attributeListCopy
        )
        if symbolPropDialogue.exec() == QDialog.DialogCode.Accepted:
            for i, item in enumerate(symbolPropDialogue.labelItemList):
                # label name is not changed.
                item.labelHeight = int(
                    float(symbolPropDialogue.labelHeightList[i].text())
                )
                item.labelAlign = symbolPropDialogue.labelAlignmentList[i].currentText()
                item.labelOrient = symbolPropDialogue.labelOrientationList[
                    i
                ].currentText()
                item.labelUse = symbolPropDialogue.labelUseList[i].currentText()
                item.labelType = symbolPropDialogue.labelTypeList[i].currentText()
                item.update(item.boundingRect())
            # create an empty attribute list. If the dialog is OK, the local attribute list
            # will be copied to the symbol attribute list.
            localAttributeList = []
            for i, item in enumerate(symbolPropDialogue.attributeNameList):
                if item.text().strip() != "":
                    localAttributeList.append(
                        symenc.symbolAttribute(
                            item.text(), symbolPropDialogue.attributeDefList[i].text()
                        )
                    )
                self.attributeList = deepcopy(localAttributeList)

    def copySelectedItems(self):
        # Only consider top-level items (same as schematic/layout editors)
        selectedItems = [item for item in self.selectedItems() if item.parentItem() is None]
        if not selectedItems:
            return
        copyShapesList = []
        factory = lj.symbolItems(self)
        for item in selectedItems:
            selectedItemJson = json.dumps(item, cls=symenc.symbolEncoder)
            itemCopyDict = json.loads(selectedItemJson)
            shape = factory.create(itemCopyDict)
            if shape is not None:
                copyShapesList.append(shape)
        if copyShapesList:
            # Add to undo stack before grouping (matches schematic/layout pattern)
            self.addListUndoStack(copyShapesList)
            # Use base-class attribute so mouseMoveEvent/mouseReleaseEvent can move
            # the group as the cursor moves and place it on mouse button release
            self.selectedItemGroup = self.createItemGroup(copyShapesList)
            self.selectedItemGroup.setSelected(True)



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#

import inspect
import json
from typing import Dict, Any

from PySide6.QtCore import QPointF

import revedaEditor.common.layoutShapes as lshp
from revedaEditor.backend.pdkLoader import importPDKModule

laylyr = importPDKModule('layoutLayers')
pcells = importPDKModule('pcells')


class layoutEncoder(json.JSONEncoder):
    def default(self, item: Any) -> Dict[str, Any]:
        if isinstance(item, lshp.layoutPcell):
            return self._encodePcell(item)
        elif isinstance(item, lshp.layoutInstance):
            return self._encodeLayoutInstance(item)
        elif isinstance(item, lshp.layoutRect):
            return self._encodeLayoutRect(item)
        elif isinstance(item, lshp.layoutPath):
            return self._encodeLayoutPath(item)
        elif isinstance(item, lshp.layoutViaArray):
            return self._encodeLayoutViaArray(item)
        elif isinstance(item, lshp.layoutPin):
            return self._encodeLayoutPin(item)
        elif isinstance(item, lshp.layoutLabel):
            return self._encodeLayoutLabel(item)
        elif isinstance(item, lshp.layoutPolygon):
            return self._encodeLayoutPolygon(item)

        return super().default(item)

    def _encodeLayoutInstance(self, item: lshp.layoutInstance) -> Dict[str, Any]:
        return {
            "type": "Inst",
            "lib": item.libraryName,
            "cell": item.cellName,
            "view": item.viewName,
            "nam": item.instanceName,
            "ic": item.counter,
            "loc": self._subtract_point(item.scenePos(), item.scene().origin),
            "ang": item.angle,
            "fl": item.flipTuple,
        }

    def _encodeLayoutRect(self, item: lshp.layoutRect) -> Dict[str, Any]:
        return {
            "type": "Rect",
            "tl": item.mapToScene(item.rect.topLeft()).toTuple(),
            "br": item.mapToScene(item.rect.bottomRight()).toTuple(),
            "ang": item.angle,
            "ln": laylyr.pdkAllLayers.index(item.layer),
            "fl": item.flipTuple,
        }

    def _encodeLayoutPath(self, item: lshp.layoutPath) -> Dict[str, Any]:
        return {
            "type": "Path",
            "dfl1": item.mapToScene(item.draftLine.p1()).toTuple(),
            "dfl2": item.mapToScene(item.draftLine.p2()).toTuple(),
            "ln": laylyr.pdkAllLayers.index(item.layer),
            "w": item.width,
            "se": item.startExtend,
            "ee": item.endExtend,
            "md": item.mode,
            "nam": item.name,
            "ang": item.angle,
            "fl": item.flipTuple,
        }

    def _encodeLayoutViaArray(self, item: lshp.layoutViaArray) -> Dict[str, Any]:
        viaDict = {
            "vdt": item.via.viaDefTuple.name,
            "st": item.via.mapToScene(item.via.start).toTuple(),
            "w": item.via.width,
            "h": item.via.height,
            "ang": item.angle,
            "fl": item.flipTuple,
        }
        return {
            "type": "Via",
            "st": item.mapToScene(item.start).toTuple(),
            "via": viaDict,
            "xs": item.xs,
            "ys": item.ys,
            "xn": item.xnum,
            "yn": item.ynum,
        }

    def _encodeLayoutPin(self, item: lshp.layoutPin) -> Dict[str, Any]:
        return {
            "type": "Pin",
            "tl": item.mapToScene(item.rect.topLeft()).toTuple(),
            "br": item.mapToScene(item.rect.bottomRight()).toTuple(),
            "pn": item.pinName,
            "pd": item.pinDir,
            "pt": item.pinType,
            "ln": laylyr.pdkAllLayers.index(item.layer),
            "ang": item.angle,
            "fl": item.flipTuple,
        }

    def _encodeLayoutLabel(self, item: lshp.layoutLabel) -> Dict[str, Any]:
        return {
            "type": "Label",
            "st": item.mapToScene(item.start).toTuple(),
            "lt": item.labelText,
            "ff": item.fontFamily,
            "fs": item.fontStyle,
            "fh": item.fontHeight,
            "la": item.labelAlign,
            "lo": item.labelOrient,
            "ln": laylyr.pdkAllLayers.index(item.layer),
            "ang": item.angle,
            "fl": item.flipTuple,
        }

    def _encodeLayoutPolygon(self, item: lshp.layoutPolygon) -> Dict[str, Any]:
        return {
            "type": "Polygon",
            "ps": [item.mapToScene(point).toTuple() for point in item.points],
            "ln": laylyr.pdkAllLayers.index(item.layer),
            "ang": item.angle,
            "fl": item.flipTuple,
        }

    def _encodePcell(self, item) -> Dict[str, Any]:
        init_args = inspect.signature(item.__class__.__init__).parameters
        args_used = [param for param in init_args if (param != "self")]
        argDict = {arg: getattr(item, arg) for arg in args_used if hasattr(item, arg)}
        return {
            "type": "Pcell",
            "lib": item.libraryName,
            "cell": item.cellName,
            "view": item.viewName,
            "nam": item.instanceName,
            "ic": item.counter,
            "loc": item.pos().toPoint().toTuple(),
            "ang": item.angle,
            "fl": item.flipTuple,
            "params": argDict,
        }

    @staticmethod
    def _subtract_point(point: QPointF, origin: QPointF) -> tuple:
        return (point - origin).toTuple()


class gdsImportEncoder(json.JSONEncoder):
    def default(self, item):
        common = {"ang": item.angle, "fl": item.flipTuple}

        match type(item):
            case lshp.layoutInstance:
                return {"type": "Inst", "lib": item.libraryName, "cell": item.cellName,
                        "view": item.viewName, "nam": item.instanceName, "ic": item.counter,
                        "loc": item.pos().toTuple(), **common}
            case lshp.layoutPath:
                return {"type": "Path",
                        "dfl1": item.mapToScene(item.draftLine.p1()).toTuple(),
                        "dfl2": item.mapToScene(item.draftLine.p2()).toTuple(),
                        "ln": laylyr.pdkAllLayers.index(item.layer), "w": item.width,
                        "se": item.startExtend, "ee": item.endExtend, "md": item.mode,
                        "nam": item.name, **common}
            case lshp.layoutViaArray:
                return {"type": "Via", "st": item.mapToScene(item.start).toTuple(),
                        "via": {"st": item.via.mapToScene(item.via.start).toTuple(),
                                "vdt": item.via.viaDefTuple.netName, "w": item.via.width,
                                "h": item.via.height, **common},
                        "xs": item.xs, "ys": item.ys, "xn": item.xnum, "yn": item.ynum}
            case lshp.layoutPin:
                return {"type": "Pin", "tl": item.mapToScene(item.rect.topLeft()).toTuple(),
                        "br": item.mapToScene(item.rect.bottomRight()).toTuple(),
                        "pn": item.pinName, "pd": item.pinDir, "pt": item.pinType,
                        "ln": laylyr.pdkAllLayers.index(item.layer), **common}
            case lshp.layoutLabel:
                return {"type": "Label", "st": item.mapToScene(item.start).toTuple(),
                        "lt": item.labelText, "ff": item.fontFamily, "fs": item.fontStyle,
                        "fh": item.fontHeight, "la": item.labelAlign,
                        "lo": item.labelOrient,
                        "ln": laylyr.pdkAllLayers.index(item.layer), **common}
            case lshp.layoutPolygon:
                return {"type": "Polygon",
                        "ps": [item.mapToScene(point).toTuple() for point in item.points],
                        "ln": laylyr.pdkAllLayers.index(item.layer), **common}



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#


from PySide6.QtCore import (
    Signal,
)
from PySide6.QtGui import (
    QPainter,
)
from PySide6.QtPrintSupport import (
    QPrintDialog,
    QPrinter,
)
from PySide6.QtWidgets import QCheckBox, QComboBox, QGridLayout, QGroupBox, QLabel


class revedaPrintDialog(QPrintDialog):
    transparencyChanged = Signal(bool)
    qualityChanged = Signal(str)

    QUALITY_OPTIONS = ["Draft", "Normal", "High"]
    DEFAULT_QUALITY = "Normal"

    def __init__(self, printer=None, parent=None):
        super().__init__(printer or QPrinter(), parent)

        self._setup_ui()

    def _setup_ui(self):
        """Setup the UI components"""
        # Create options group box
        optionsGroup = QGroupBox("Print Options", self)
        optionsLayout = QGridLayout(optionsGroup)

        # Setup transparency checkbox
        self._setup_transparency(optionsLayout)

        # Setup quality options
        self._setup_quality_options(optionsLayout)

        # Add the options group to the dialog
        self.layout().addWidget(optionsGroup)

    def _setup_transparency(self, layout):
        """Setup transparency checkbox"""
        self.transparentCheck = QCheckBox("Transparent Background")
        self.transparentCheck.setChecked(True)
        # Connect using the clicked signal instead of stateChanged
        self.transparentCheck.clicked.connect(self.transparencyChanged.emit)
        layout.addWidget(self.transparentCheck, 0, 0)

    def _setup_quality_options(self, layout):
        """Setup quality combo box and label"""
        qualityLabel = QLabel("Print Quality:")
        self.qualityCombo = QComboBox()
        self.qualityCombo.addItems(self.QUALITY_OPTIONS)
        self.qualityCombo.setCurrentText(self.DEFAULT_QUALITY)
        self.qualityCombo.currentTextChanged.connect(self.qualityChanged.emit)

        layout.addWidget(qualityLabel, 1, 0)
        layout.addWidget(self.qualityCombo, 1, 1)

    def isTransparent(self):
        """Return whether transparent background is selected"""
        return self.transparentCheck.isChecked()

    def getPrintQuality(self):
        """Return selected print quality"""
        quality_map = {
            "High": QPrinter.HighResolution,
            "Draft": QPrinter.ScreenResolution,
            "Normal": QPrinter.StandardResolution
        }
        return quality_map.get(self.qualityCombo.currentText(), QPrinter.StandardResolution)

    def getRenderHints(self):
        """Return appropriate render hints based on quality setting"""
        current_quality = self.qualityCombo.currentText()
        hints = QPainter.TextAntialiasing

        if current_quality != "Draft":
            hints |= QPainter.Antialiasing | QPainter.SmoothPixmapTransform

        if current_quality == "High":
            hints |= QPainter.HighQualityAntialiasing

        return hints



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#

from __future__ import annotations

# import numpy as np
from PySide6.QtCore import (
    Qt,
)
from PySide6.QtWidgets import (
    QDialog,
    QGridLayout,
    QToolBar,
)

import revedaEditor.backend.dataDefinitions as ddef
import revedaEditor.backend.libBackEnd as libb
import revedaEditor.backend.libraryModelView as lmview
import revedaEditor.gui.editorViews as edv
import revedaEditor.gui.editorWindow as edw
import revedaEditor.gui.propertyDialogues as pdlg
import revedaEditor.scenes.symbolScene as symscn


# from hashlib import new


class symbolEditor(edw.editorWindow):
    def __init__(
            self,
            viewItem: libb.viewItem,
            libraryDict: dict,
            libraryView: lmview.BaseDesignLibrariesView,
    ):
        super().__init__(viewItem, libraryDict, libraryView)
        self.setWindowTitle(f"Symbol Editor - {self.cellName} - {self.viewName}")
        self._symbolContextMenu()

    def init_UI(self):
        super().init_UI()
        # create container to position all widgets
        self.centralW = symbolContainer(self)
        self.setCentralWidget(self.centralW)

    # def _createActions(self):
    #     super()._createActions()

    def _createShortcuts(self):
        super()._createShortcuts()
        self.stretchAction.setShortcut(Qt.Key.Key_S)
        self.createRectAction.setShortcut(Qt.Key.Key_R)
        self.createLineAction.setShortcut(Qt.Key.Key_W)
        self.createLabelAction.setShortcut(Qt.Key.Key_L)
        self.createPinAction.setShortcut(Qt.Key.Key_P)

    def _createToolBars(self):  # redefine the toolbar in the editorWindow class
        super()._createToolBars()
        self.symbolToolbar = QToolBar("Symbol Toolbar", self)
        self.addToolBar(self.symbolToolbar)
        self.symbolToolbar.addAction(self.createLineAction)
        self.symbolToolbar.addAction(self.createRectAction)
        self.symbolToolbar.addAction(self.createPolygonAction)
        self.symbolToolbar.addAction(self.createCircleAction)
        self.symbolToolbar.addAction(self.createArcAction)
        self.symbolToolbar.addAction(self.createLabelAction)
        self.symbolToolbar.addAction(self.createPinAction)

    def _addActions(self):
        super()._addActions()
        self.menuEdit.addAction(self.stretchAction)
        self.menuEdit.addAction(self.viewPropAction)
        self.menuCreate.addAction(self.createLineAction)
        self.menuCreate.addAction(self.createRectAction)
        self.menuCreate.addAction(self.createPolygonAction)
        self.menuCreate.addAction(self.createCircleAction)
        self.menuCreate.addAction(self.createArcAction)
        self.menuCreate.addAction(self.createLabelAction)
        self.menuCreate.addAction(self.createPinAction)

        if hasattr(self._app, 'pluginsObj') and hasattr(self._app.pluginsObj, 'applyPluginMenus'):
            self._app.pluginsObj.applyPluginMenus(self)

    def _createTriggers(self):
        super()._createTriggers()
        self.createLineAction.triggered.connect(self.createLineClick)
        self.createRectAction.triggered.connect(self.createRectClick)
        self.createPolygonAction.triggered.connect(self.createPolyClick)
        self.createArcAction.triggered.connect(self.createArcClick)
        self.createCircleAction.triggered.connect(self.createCircleClick)
        self.createLabelAction.triggered.connect(self.createLabelClick)
        self.createPinAction.triggered.connect(self.createPinClick)
        self.objPropAction.triggered.connect(self.objPropClick)
        self.deleteAction.triggered.connect(self.deleteClick)
        self.viewPropAction.triggered.connect(self.viewPropClick)

    def _symbolContextMenu(self):
        super()._editorContextMenu()
        self.centralW.scene.itemContextMenu.addAction(self.stretchAction)

    def objPropClick(self):
        self.centralW.scene.itemProperties()

    def checkSaveCell(self):
        self.centralW.scene.saveSymbolCell(self.file)
        if self.parentEditor:
            self.parentEditor.centralW.scene.reloadScene()

    def saveCell(self):
        self.centralW.scene.saveSymbolCell(self.file)

    def createRectClick(self, s):
        self.centralW.scene.editModes.setMode("drawRect")
        self.messageLine.setText("Press left mouse button for the first point.")

    def createLineClick(self, s):
        self.centralW.scene.editModes.setMode("drawLine")
        self.messageLine.setText("Press left mouse button for the first point.")

    def createPolyClick(self, s):
        self.centralW.scene.editModes.setMode("drawPolygon")
        self.messageLine.setText("Press left mouse button for the first point.")

    def createArcClick(self, s):
        self.centralW.scene.editModes.setMode("drawArc")
        self.messageLine.setText("Press left mouse button for the first point.")

    def createCircleClick(self, s):
        self.centralW.scene.editModes.setMode("drawCircle")
        self.messageLine.setText("Press left mouse button for the centre point.")

    def createPinClick(self, s):
        createPinDlg = pdlg.createPinDialog(self)
        if createPinDlg.exec() == QDialog.DialogCode.Accepted:
            modeList = [False for _ in range(8)]
            modeList[0] = True
            self.centralW.scene.pinName = createPinDlg.pinName.text()
            self.centralW.scene.pinType = createPinDlg.pinType.currentText()
            self.centralW.scene.pinDir = createPinDlg.pinDir.currentText()
            self.centralW.scene.editModes.setMode("drawPin")
            self.messageLine.setText("Place pin on the symbol.")

    def rotateItemClick(self, s):
        self.centralW.scene.editModes.setMode("rotateItem")
        self.messageLine.setText("Click on an item to rotate CW 90 degrees.")

    def copyClick(self, s):
        self.centralW.scene.editModes.setMode("copyItem")
        self.centralW.scene.copySelectedItems()
        self.messageLine.setText("Copying selected items")

    def viewPropClick(self, s):
        self.centralW.scene.editModes.setMode("selectItem")
        self.centralW.scene.viewSymbolProperties()

    def loadSymbol(self):
        """
        symbol is loaded to the scene.
        """
        try:
            self.logger.info(f'Loading symbol from {self.cellName} - {self.viewName}')
            self.centralW.scene.loadDesign(self.file)
            viewNameTuple = ddef.viewNameTuple(self.libItem.libraryName, self.cellItem.cellName,
                                               self.viewName)
            self.appMainW.openViews[viewNameTuple] = self
        except Exception as e:
            self.logger.error(f"Error during loading symbol for {self.cellName}: {e}")

    def createLabelClick(self):
        createLabelDlg = pdlg.createSymbolLabelDialog(self)
        self.messageLine.setText("Place a label")
        createLabelDlg.labelHeightEdit.setText("12")
        if createLabelDlg.exec() == QDialog.DialogCode.Accepted:
            self.centralW.scene.editModes.setMode("addLabel")
            # directly setting scene class attributes here to pass the information.
            self.centralW.scene.labelDefinition = createLabelDlg.labelDefinition.text()
            self.centralW.scene.labelHeight = (
                createLabelDlg.labelHeightEdit.text().strip()
            )
            self.centralW.scene.labelAlignment = (
                createLabelDlg.labelAlignCombo.currentText()
            )
            self.centralW.scene.labelOrient = (
                createLabelDlg.labelOrientCombo.currentText()
            )
            self.centralW.scene.labelUse = createLabelDlg.labelUseCombo.currentText()
            self.centralW.scene.labelOpaque = (
                    createLabelDlg.labelVisiCombo.currentText() == "Yes"
            )
            self.centralW.scene.labelType = "Normal"  # default button
            if createLabelDlg.normalType.isChecked():
                self.centralW.scene.labelType = "Normal"
            elif createLabelDlg.NLPType.isChecked():
                self.centralW.scene.labelType = "NLPLabel"
            elif createLabelDlg.pyLType.isChecked():
                self.centralW.scene.labelType = "PyLabel"

    def closeEvent(self, event):
        """
        Closes the application.
        """
        try:
            self.centralW.scene.saveSymbolCell(self.file)
            cellViewNameTuple = ddef.viewNameTuple(self.libName, self.cellName,
                                                   self.viewName)
            self.appMainW.openViews.pop(cellViewNameTuple, None)
        except Exception as e:
            self.appMainW.logger.error(
                f"Error in closing symbol window for {self.cellName}: {e}")
        finally:
            event.accept()
            super().closeEvent(event)


class symbolContainer(edw.editorContainer):
    def __init__(self, parent):
        super().__init__(parent=parent)
        self.editorWindow = parent
        self.scene = symscn.symbolScene(self)
        self.view = edv.symbolView(self.scene, self)
        self.init_UI()

    def init_UI(self):
        # layout statements, using a grid layout
        gLayout = QGridLayout()
        gLayout.setSpacing(10)
        gLayout.addWidget(self.view, 0, 0)
        gLayout.setColumnStretch(0, 5)
        gLayout.setRowStretch(0, 6)
        self.setLayout(gLayout)



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#   #
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#   #
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#   #
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)

from typing import Union

from PySide6.QtCore import (
    Qt,
)
from PySide6.QtGui import QStandardItemModel

import revedaEditor.backend.libBackEnd as scb


def getLibItem(
        libraryModel: QStandardItemModel, libName: str
) -> Union[scb.libraryItem, None]:
    return next((item for item in libraryModel.findItems(libName)
                 if item.data(Qt.ItemDataRole.UserRole + 1) == "library"), None)


def getCellItem(
        libItem: Union[scb.libraryItem, None], cellNameInp: str
) -> Union[scb.cellItem, None]:
    if libItem is None:
        return None
    return next((libItem.child(i) for i in range(libItem.rowCount())
                 if libItem.child(i) and libItem.child(i).cellName == cellNameInp), None)


def getViewItem(cellItem: Union[scb.cellItem, None], viewNameInp: str) -> Union[
    scb.viewItem, None]:
    if cellItem is None:
        return None
    return next((cellItem.child(i) for i in range(cellItem.rowCount())
                 if cellItem.child(i) and cellItem.child(i).text() == viewNameInp), None)


def findViewItem(libraryModel, libName: str, cellName: str, viewName: str) -> Union[
    scb.viewItem, None]:
    libItem = getLibItem(libraryModel, libName)
    if libItem:
        cellItem = getCellItem(libItem, cellName)
        if cellItem:
            return getViewItem(cellItem, viewName)
    return None
